from time import perf_counter
from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.visualization.heatmap_fast import sinr_heatmap_2d_fast

def big_heatmap(res=600):
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(1000,1000,30)))
    twin.set_propagation_model(FreeSpacePathLoss())
    net = Network()
    net.add_node(WirelessNode("gNB-1",(300,300,10),TransceiverProperties(32,5)))
    net.add_node(WirelessNode("gNB-2",(700,700,10),TransceiverProperties(32,5)))
    twin.network = net

    t0 = perf_counter()
    _, Z = sinr_heatmap_2d_fast(
        twin, tx_id="gNB-1", interferer_ids=["gNB-2"],
        xlim=(0,1000), ylim=(0,1000), resolution=res, show=False
    )
    dt = perf_counter()-t0
    print(f"Heatmap {res}x{res} -> {Z.shape} in {dt:.3f}s")

if __name__ == "__main__":
    big_heatmap(300)
    big_heatmap(500)
    big_heatmap(700)




from __future__ import annotations
from dataclasses import dataclass
from typing import Tuple, Optional
from ..core.node import WirelessNode

@dataclass
class ConstantVelocity:
    velocity_mps: Tuple[float, float, float] = (0.0, 0.0, 0.0)
    start_time: float = 0.0
    _last_t: Optional[float] = None
    def update(self, node: WirelessNode, timestamp: float) -> None:
        if self._last_t is None:
            self._last_t = timestamp
            return
        dt = max(0.0, timestamp - self._last_t)
        self._last_t = timestamp
        x, y, z = node.position
        vx, vy, vz = self.velocity_mps
        node.move_to((x + vx * dt, y + vy * dt, z + vz * dt))



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, Tuple
import math
from ..core.node import WirelessNode
from .environment import Environment
from .propagation import PropagationModel

@dataclass
class SmartRISPanel:
    """
    Toy RIS with a steerable mainlobe. Approximates array gain:
      mainlobe_gain_db  ~= 20*log10(N)
      sidelobe_gain_db  ~= mainlobe - 13 dB (typical)
    """
    position: Tuple[float, float, float]
    element_count: int = 64
    mainlobe_gain_db: Optional[float] = None
    sidelobe_gain_db: Optional[float] = None

    def __post_init__(self):
        if self.mainlobe_gain_db is None:
            self.mainlobe_gain_db = 20.0 * math.log10(max(self.element_count, 1))
        if self.sidelobe_gain_db is None:
            self.sidelobe_gain_db = self.mainlobe_gain_db - 13.0

class RISBeamModel(PropagationModel):
    """
    Wraps a base PropagationModel. For a configured target UE (by id), the RIS
    contributes mainlobe gain on the two-hop path; others see sidelobe gain.
    Effective path loss = min( direct, (tx->RIS + RIS->rx - gain + extra_loss_db) ).
    """
    def __init__(self, base: PropagationModel, ris: SmartRISPanel, extra_loss_db: float = 3.0) -> None:
        self.base = base
        self.ris = ris
        self.extra_loss_db = extra_loss_db
        self._beam_target_id: Optional[str] = None

    def set_beam(self, rx_node_id: Optional[str]) -> None:
        """Point mainlobe toward rx_node_id (or None for no specific target)."""
        self._beam_target_id = rx_node_id

    def calculate_path_loss(self, tx: WirelessNode, rx: WirelessNode, environment: Environment) -> float:
        direct = self.base.calculate_path_loss(tx, rx, environment)
        tmp1 = WirelessNode("__tmp_tx2ris__", position=self.ris.position)
        tmp2 = WirelessNode("__tmp_ris2rx__", position=self.ris.position)
        pl1 = self.base.calculate_path_loss(tx, tmp1, environment)
        pl2 = self.base.calculate_path_loss(tmp2, rx, environment)
        gain = self.ris.mainlobe_gain_db if (self._beam_target_id == rx.node_id) else self.ris.sidelobe_gain_db
        via = pl1 + pl2 - gain + self.extra_loss_db
        return min(direct, via)



from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.optimization.objective import SumThroughputObjective
from pytwinnet.optimization.placement_parallel import ParallelPlacementGrid

def main():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(300, 300, 30)))
    twin.set_propagation_model(FreeSpacePathLoss())
    net = Network()
    net.add_node(WirelessNode("gNB-1", position=(20, 20, 10),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=32, antenna_gain_dbi=5),
                              metadata={"role":"gNB"}))
    for i, pos in enumerate([(60,80,1.5),(120,90,1.5),(220,150,1.5),(260,240,1.5)]):
        net.add_node(WirelessNode(f"UE-{i+1}", position=pos))
    twin.network = net

    obj = SumThroughputObjective(tx_id="gNB-1", efficiency=0.75)
    grid_x = [50, 100, 150, 200, 250]
    grid_y = [50, 100, 150, 200, 250]

    optimizer = ParallelPlacementGrid(grid_x=grid_x, grid_y=grid_y, fixed_z=10.0, max_workers=0)
    res = optimizer.optimize(twin, obj, node_ids=["gNB-1"])
    print("Best:", res["best_positions"], "score=", f"{res['best_score']:.2e}", "evals=", res["evaluations"])

if __name__ == "__main__":
    main()



from __future__ import annotations
from typing import Dict, List
from ..core.digital_twin import DigitalTwin
from ..physics.link_budget import rx_power_dbm, noise_power_dbm, sinr_db, shannon_throughput_bps

def _est_throughput_bps(twin: DigitalTwin, tx_id: str, ue_id: str,
                        bandwidth_hz: float, noise_figure_db: float) -> float:
    pm = twin.propagation_model; env = twin.environment
    tx = twin.network.get_node_by_id(tx_id); ue = twin.network.get_node_by_id(ue_id)
    pl_db = pm.calculate_path_loss(tx, ue, env)
    prx_dbm = rx_power_dbm(
        tx.transceiver_properties.transmit_power_dbm,
        tx.transceiver_properties.antenna_gain_dbi,
        ue.transceiver_properties.antenna_gain_dbi,
        pl_db,
    )
    n_dbm = noise_power_dbm(290.0, bandwidth_hz, noise_figure_db)
    s = sinr_db(prx_dbm, interferers_dbm=None, noise_dbm=n_dbm)
    return shannon_throughput_bps(bandwidth_hz, s, efficiency=0.75)

def proportional_fair_allocation(
    twin: DigitalTwin,
    association: Dict[str, str],
    rb_count: int = 50,
    rb_bandwidth_hz: float = 180e3,
    noise_figure_db: float = 7.0,
    avg_throughput_bps: Dict[str, float] | None = None,
) -> Dict[str, List[str]]:
    """Greedy PF: per TX, for each RB pick UE maximizing inst_rate / avg_rate."""
    if avg_throughput_bps is None:
        avg_throughput_bps = {}
    eps = 1e-6

    # group UEs per TX
    per_tx: Dict[str, List[str]] = {}
    for ue_id, tx_id in association.items():
        per_tx.setdefault(tx_id, []).append(ue_id)

    schedule: Dict[str, List[str]] = {tx_id: [] for tx_id in per_tx.keys()}
    for tx_id, ues in per_tx.items():
        if not ues:
            continue
        for _ in range(rb_count):
            best_ue, best_metric = None, float("-inf")
            for ue_id in ues:
                inst = _est_throughput_bps(twin, tx_id, ue_id, rb_bandwidth_hz, noise_figure_db)
                avg = avg_throughput_bps.get(ue_id, inst)  # cold-start
                metric = inst / max(avg, eps)
                if metric > best_metric:
                    best_metric, best_ue = metric, ue_id
            schedule[tx_id].append(best_ue)
    return schedule




import os, sys
sys.path.insert(0, os.path.abspath('..'))
project = 'PyTwinNet'
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.napoleon']
templates_path = ['_templates']
exclude_patterns = []
html_theme = 'sphinx_rtd_theme'




from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.core.node import WirelessNode

def test_fspl_increases_with_distance():
    pm = FreeSpacePathLoss()
    env = Environment()
    tx = WirelessNode("tx", position=(0,0,0))
    rx1 = WirelessNode("rx1", position=(1,0,0))
    rx2 = WirelessNode("rx2", position=(10,0,0))
    pl1 = pm.calculate_path_loss(tx, rx1, env)
    pl2 = pm.calculate_path_loss(tx, rx2, env)
    assert pl2 > pl1




from __future__ import annotations
import math
from typing import Iterable, Optional

def db_to_linear(db: float) -> float:
    return 10 ** (db / 10.0)

def linear_to_db(x: float) -> float:
    return 10.0 * math.log10(max(x, 1e-30))

def rx_power_dbm(tx_power_dbm: float, tx_gain_dbi: float, rx_gain_dbi: float, path_loss_db: float) -> float:
    return tx_power_dbm + tx_gain_dbi + rx_gain_dbi - path_loss_db

def noise_power_dbm(temperature_k: float, bandwidth_hz: float, noise_figure_db: float = 0.0) -> float:
    ref_dbm_per_hz = -174.0 + linear_to_db(temperature_k / 290.0)
    return ref_dbm_per_hz + 10.0 * math.log10(max(bandwidth_hz, 1.0)) + noise_figure_db

def sinr_db(signal_dbm: float, interferers_dbm: Optional[Iterable[float]] = None, noise_dbm: float = -100.0) -> float:
    def db2mw(v): return 10 ** (v / 10.0)
    sig_mw = db2mw(signal_dbm)
    int_mw = sum(db2mw(p) for p in (interferers_dbm or []))
    noi_mw = db2mw(noise_dbm)
    return 10.0 * math.log10(sig_mw / (int_mw + noi_mw))

def shannon_throughput_bps(bandwidth_hz: float, sinr_db_value: float, efficiency: float = 1.0) -> float:
    sinr = 10 ** (sinr_db_value / 10.0)
    return efficiency * bandwidth_hz * math.log2(1.0 + sinr)



from __future__ import annotations
from dataclasses import dataclass
from typing import Tuple
from ..core.node import WirelessNode
from .environment import Environment
from .propagation import PropagationModel

@dataclass
class RISPanel:
    position: Tuple[float, float, float]
    gain_db: float = 10.0  # toy reflective gain

class RISAugmentedModel(PropagationModel):
    """
    Wrap a base PropagationModel and return the min of:
    - direct path loss
    - two-hop path loss via RIS: PL(tx->RIS) + PL(RIS->rx) - RIS_gain + extra_loss
    """
    def __init__(self, base: PropagationModel, ris: RISPanel, extra_loss_db: float = 3.0) -> None:
        self.base = base
        self.ris = ris
        self.extra_loss_db = extra_loss_db

    def calculate_path_loss(self, tx: WirelessNode, rx: WirelessNode, environment: Environment) -> float:
        direct = self.base.calculate_path_loss(tx, rx, environment)
        tx2ris = WirelessNode("__tmp_tx2ris__", position=self.ris.position)
        ris2rx = WirelessNode("__tmp_ris2rx__", position=self.ris.position)
        pl1 = self.base.calculate_path_loss(tx, tx2ris, environment)
        pl2 = self.base.calculate_path_loss(ris2rx, rx, environment)
        via_ris = pl1 + pl2 - self.ris.gain_db + self.extra_loss_db
        return min(direct, via_ris)



"""
Ingestion + Optimization demo for PyTwinNet.

- Build a twin with gNB + 2 UEs
- Apply a few mock real-time updates (node movement)
- Generate traffic to create a naive "throughput" signal
- Run SimpleGreedyOptimizer to improve the (toy) throughput objective
- Print path loss and final node states
"""

from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.simulation import Scenario, MoveNodeEvent, Simulator, TrafficGenerationEvent
from pytwinnet.ingestion import MockDataSource, RealTimeMonitor
from pytwinnet.optimization import SimpleGreedyOptimizer
from pytwinnet.optimization.objective import MaximizeThroughput


def main():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(200.0, 200.0, 30.0)))
    twin.set_propagation_model(FreeSpacePathLoss())

    # --- Build a simple network ---
    net = Network()
    net.add_node(WirelessNode(node_id="gNB-1", position=(0.0, 0.0, 10.0),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=30.0)))
    net.add_node(WirelessNode(node_id="UE-1", position=(10.0, 0.0, 1.5)))
    net.add_node(WirelessNode(node_id="UE-2", position=(0.0, 20.0, 1.5)))
    twin.network = net

    # --- Ingestion demo: move UEs a bit using mock updates ---
    source = MockDataSource(node_ids=["UE-1", "UE-2"], step_size_m=2.0, seed=42)
    source.connect()
    monitor = RealTimeMonitor(twin=twin, source=source)
    applied = monitor.poll_once()  # apply one batch of random small movements
    print(f"[Ingestion] Applied {applied} mock updates.")
    for n in twin.network.list_nodes():
        print(f"  {n.node_id} -> position={n.position}")

    # --- Generate some traffic (toy) ---
    scen = Scenario(duration_s=5.0)
    scen.add_event(TrafficGenerationEvent(timestamp=1.0, source_node="gNB-1", dest_node="UE-1", data_rate_mbps=5.0))
    scen.add_event(TrafficGenerationEvent(timestamp=2.0, source_node="gNB-1", dest_node="UE-2", data_rate_mbps=2.5))
    Simulator(twin).run(scenario=scen, copy_twin=False)  # mutate twin directly for demonstration

    # --- Compute path loss between gNB and each UE (FSPL) ---
    fspl = twin.propagation_model
    env = twin.environment
    gnb = twin.network.get_node_by_id("gNB-1")
    for ue_id in ["UE-1", "UE-2"]:
        ue = twin.network.get_node_by_id(ue_id)
        pl_db = fspl.calculate_path_loss(gnb, ue, env)
        print(f"[Physics] FSPL(gNB-1 -> {ue_id}) = {pl_db:.2f} dB")

    # --- Optimization: greedy improve 'MaximizeThroughput' (toy) ---
    opt = SimpleGreedyOptimizer(step_db=1.0, max_power_dbm=33.0, iterations=5)
    obj = MaximizeThroughput()
    summary = opt.optimize(twin, obj)
    print(f"[Optimization] best_score={summary['best_score']:.2f}, iterations={summary['iterations']}")
    print("  Power updates history (node_id, score):", summary["metadata"]["history"])

    # --- Final state ---
    for node in twin.network.list_nodes():
        print(f"{node.node_id}: position={node.position}, "
              f"tx_power={node.transceiver_properties.transmit_power_dbm:.1f} dBm, "
              f"recv_traffic={node.metadata.get('received_traffic_mbps', 0.0)} Mbps")


if __name__ == "__main__":
    main()




from __future__ import annotations
import random
from dataclasses import dataclass
from typing import Iterable, List, Tuple
from .base import DataSource

@dataclass
class MockNodeUpdate:
    node_id: str
    new_position: Tuple[float, float, float]

class MockDataSource(DataSource):
    def __init__(self, node_ids: List[str], step_size_m: float = 1.0, seed: int = 0) -> None:
        self.node_ids = node_ids
        self.step = step_size_m
        self.rng = random.Random(seed)
    def connect(self) -> None:
        return None
    def read_data(self) -> Iterable[MockNodeUpdate]:
        for nid in self.node_ids:
            dx = (self.rng.random() - 0.5) * 2.0 * self.step
            dy = (self.rng.random() - 0.5) * 2.0 * self.step
            dz = 0.0
            yield MockNodeUpdate(node_id=nid, new_position=(dx, dy, dz))




from __future__ import annotations
from dataclasses import dataclass
from typing import List
import random
@dataclass
class PoissonTraffic:
    rate_lambda: float
    seed: int = 0
    def arrivals(self, duration_s: float) -> List[float]:
        rng = random.Random(self.seed)
        t = 0.0; times = []
        while t <= duration_s:
            dt = rng.expovariate(self.rate_lambda) if self.rate_lambda > 0 else duration_s + 1
            t += dt
            if t <= duration_s:
                times.append(t)
        return times




from __future__ import annotations
from dataclasses import dataclass, field
from typing import Dict, Iterable, Optional, List
from .node import WirelessNode

@dataclass
class Network:
    nodes: Dict[str, WirelessNode] = field(default_factory=dict)
    def add_node(self, node: WirelessNode) -> None:
        self.nodes[node.node_id] = node
    def remove_node(self, node_id: str) -> None:
        if node_id in self.nodes:
            del self.nodes[node_id]
    def get_node_by_id(self, node_id: str) -> Optional[WirelessNode]:
        return self.nodes.get(node_id)
    def list_nodes(self) -> List[WirelessNode]:
        return list(self.nodes.values())
    def __iter__(self) -> Iterable[WirelessNode]:
        return iter(self.nodes.values())



from __future__ import annotations
import numpy as np
from typing import Dict, List, Tuple
from ..core.digital_twin import DigitalTwin
from ..core.node import WirelessNode
from ..physics.propagation import PropagationModel
from ..physics.environment import Environment
from .vectorized import fspl_matrix_db, rsrp_matrix_dbm

def _stack_positions(nodes: List[WirelessNode]) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    xyz = np.array([n.position for n in nodes], dtype=float)   # (N,3)
    pw = np.array([n.transceiver_properties.transmit_power_dbm for n in nodes], dtype=float)
    g  = np.array([n.transceiver_properties.antenna_gain_dbi for n in nodes], dtype=float)
    return xyz, pw, g

def max_rsrp_association_vectorized(
    twin: DigitalTwin, tx_ids: List[str], ue_ids: List[str]
) -> Dict[str, str]:
    """
    Vectorized max-RSRP association. Much faster than per-pair loops for large sets.
    """
    pm: PropagationModel = twin.propagation_model
    env: Environment = twin.environment
    assert pm is not None and env is not None

    txs = [twin.network.get_node_by_id(t) for t in tx_ids]
    ues = [twin.network.get_node_by_id(u) for u in ue_ids]
    txs = [t for t in txs if t is not None]
    ues = [u for u in ues if u is not None]
    if not txs or not ues:
        return {}

    tx_xyz, tx_dbm, gt_dbi = _stack_positions(txs)
    ue_xyz, _, gr_dbi = _stack_positions(ues)
    f_hz = txs[0].transceiver_properties.carrier_frequency_hz  # assume same for now

    # Build PL matrix via underlying scalar model by matching FSPL formula for speed.
    # If pm is not FSPL, we fall back to slow path.
    from ..physics.propagation import FreeSpacePathLoss
    if isinstance(pm, FreeSpacePathLoss):
        pl_db = fspl_matrix_db(tx_xyz, ue_xyz, f_hz)  # (T,R)
    else:
        # Slow fallback (still vectorizable with numba in future)
        import numpy as np
        pl_db = np.zeros((len(txs), len(ues)), dtype=float)
        for i, tx in enumerate(txs):
            for j, rx in enumerate(ues):
                pl_db[i, j] = pm.calculate_path_loss(tx, rx, env)

    rsrp_dbm = rsrp_matrix_dbm(tx_dbm, gt_dbi, gr_dbi, pl_db)  # (T,R)
    best_tx_idx = rsrp_dbm.argmax(axis=0)                      # (R,)
    mapping = {ues[j].node_id: txs[i].node_id for j, i in enumerate(best_tx_idx)}
    return mapping



import numpy as np
from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.accelerate import fspl_matrix_db, rsrp_matrix_dbm, max_rsrp_association_vectorized

def main():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(1000, 1000, 30)))
    twin.set_propagation_model(FreeSpacePathLoss())

    net = Network()
    # 10 gNBs
    for i in range(10):
        net.add_node(WirelessNode(
            node_id=f"gNB-{i+1}",
            position=(100 + 80*i, 100 + 30*i, 10.0),
            transceiver_properties=TransceiverProperties(transmit_power_dbm=32.0, antenna_gain_dbi=5.0),
            metadata={"role": "gNB"},
        ))
    # 200 UEs on a grid
    for r in range(20):
        for c in range(10):
            idx = r*10 + c + 1
            net.add_node(WirelessNode(node_id=f"UE-{idx}", position=(20 + 45*c, 20 + 40*r, 1.5)))
    twin.network = net

    tx_ids = [f"gNB-{i+1}" for i in range(10)]
    ue_ids = [f"UE-{i+1}" for i in range(200)]

    assoc = max_rsrp_association_vectorized(twin, tx_ids, ue_ids)
    # Show first 10 associations
    print({k: assoc[k] for k in list(assoc)[:10]})

if __name__ == "__main__":
    main()



from __future__ import annotations
import numpy as np
from typing import Tuple, Optional


try:
    from .numba_kernels import HAVE_NUMBA, fspl_matrix_db_numba, sinr_db_from_rsrp_matrix_numba
except Exception:  # pragma: no cover
    HAVE_NUMBA = False
    fspl_matrix_db_numba = None
    sinr_db_from_rsrp_matrix_numba = None
# ---------- Geometry / FSPL ----------

def _pairwise_dist(tx_xyz: np.ndarray, rx_xyz: np.ndarray) -> np.ndarray:
    """
    tx_xyz: (T,3), rx_xyz: (R,3) -> (T,R) 3D distances (meters)
    """
    d = tx_xyz[:, None, :] - rx_xyz[None, :, :]
    return np.linalg.norm(d, axis=2)

def fspl_matrix_db(
    tx_xyz: np.ndarray, rx_xyz: np.ndarray, f_hz: np.ndarray | float
) -> np.ndarray:
    """
    Vectorized Free-Space Path Loss in dB: (T,R)
    If f_hz is scalar and Numba is available -> JIT kernel.
    """
    if np.isscalar(f_hz):
        f_mhz = float(f_hz) / 1e6
        if HAVE_NUMBA and fspl_matrix_db_numba is not None:
            # Numba prefers float64
            return fspl_matrix_db_numba(
                np.asarray(tx_xyz, dtype=np.float64),
                np.asarray(rx_xyz, dtype=np.float64),
                float(f_mhz),
            )
        # NumPy fallback
        d = tx_xyz[:, None, :] - rx_xyz[None, :, :]
        d_m = np.linalg.norm(d, axis=2)
        d_km = np.maximum(d_m, 1e-3) / 1e3
        return 20.0 * np.log10(d_km) + 20.0 * np.log10(f_mhz) + 32.44
    # Vector f_hz: use NumPy path
    d = tx_xyz[:, None, :] - rx_xyz[None, :, :]
    d_m = np.linalg.norm(d, axis=2)
    d_km = np.maximum(d_m, 1e-3) / 1e3
    f_mhz = np.asarray(f_hz, dtype=float) / 1e6  # (T,)
    return 20.0 * np.log10(d_km) + 20.0 * np.log10(f_mhz[:, None]) + 32.44


# ---------- Link budget (vectorized) ----------

def rsrp_matrix_dbm(
    tx_dbm: np.ndarray, gt_dbi: np.ndarray, gr_dbi: np.ndarray, pl_db: np.ndarray
) -> np.ndarray:
    """
    RSRP (received power) in dBm for all TX-RX pairs.
    tx_dbm, gt_dbi: (T,), gr_dbi: (R,), pl_db: (T,R)
    -> (T,R)
    """
    return tx_dbm[:, None] + gt_dbi[:, None] + gr_dbi[None, :] - pl_db

def noise_dbm_vector(
    bandwidth_hz: np.ndarray | float,
    temperature_k: float = 290.0,
    noise_figure_db: np.ndarray | float = 0.0,
) -> np.ndarray:
    """
    Vectorized thermal noise power (dBm).
    bandwidth_hz: scalar or (R,), noise_figure_db: scalar or (R,)
    -> (R,)
    """
    # -174 dBm/Hz at 290K baseline
    # Adjust for temperature linearly in power domain -> +10*log10(T/290)
    temp_adj_db = 10.0 * np.log10(np.asarray(temperature_k, dtype=float) / 290.0)
    bw = np.asarray(bandwidth_hz, dtype=float)
    nf = np.asarray(noise_figure_db, dtype=float)
    return -174.0 + temp_adj_db + 10.0 * np.log10(np.maximum(bw, 1.0)) + nf

def sinr_db_from_rsrp_matrix(
    rsrp_dbm: np.ndarray,  # (T,R)
    serving_tx_idx: np.ndarray,  # (R,)
    noise_dbm: np.ndarray,       # (R,)
) -> np.ndarray:
    if HAVE_NUMBA and sinr_db_from_rsrp_matrix_numba is not None:
        return sinr_db_from_rsrp_matrix_numba(
            np.asarray(rsrp_dbm, dtype=np.float64),
            np.asarray(serving_tx_idx, dtype=np.int64),
            np.asarray(noise_dbm, dtype=np.float64),
        )
    # NumPy fallback
    p_mw = 10 ** (rsrp_dbm / 10.0)
    total_mw = p_mw.sum(axis=0)
    sig_mw = p_mw[serving_tx_idx, np.arange(p_mw.shape[1])]
    noi_mw = 10 ** (noise_dbm / 10.0)
    sinr_lin = sig_mw / np.maximum(total_mw - sig_mw + noi_mw, 1e-30)
    return 10.0 * np.log10(sinr_lin)


def shannon_throughput_bps_vector(
    bandwidth_hz: np.ndarray | float,
    sinr_db: np.ndarray,              # (R,)
    efficiency: float = 1.0,
) -> np.ndarray:
    """
    Vectorized Shannon throughput per RX (bps).
    """
    bw = np.asarray(bandwidth_hz, dtype=float)
    sinr_lin = 10 ** (sinr_db / 10.0)
    return efficiency * bw * np.log2(1.0 + sinr_lin)




from __future__ import annotations
from typing import List
from .network import Network

def list_gnbs(network: Network) -> List[str]:
    out = []
    for n in network:
        role = str(n.metadata.get("role", "")).lower()
        if "gnb" in role or "bs" in role:
            out.append(n.node_id)
    return out



from __future__ import annotations
import csv
from dataclasses import dataclass
from typing import Dict, Iterable, Iterator
from .base import DataSource
from ..core.digital_twin import DigitalTwin

@dataclass
class CSVPositionDataSource(DataSource):
    """CSV source for node positions.

    CSV (header required):
        timestamp_s,node_id,x,y,z
    """
    path: str

    def connect(self) -> None:
        return

    def read_data(self) -> Iterator[Dict]:
        with open(self.path, newline="", encoding="utf-8") as f:
            r = csv.DictReader(f)
            for row in r:
                yield {
                    "timestamp_s": float(row["timestamp_s"]),
                    "node_id": row["node_id"],
                    "position": (float(row["x"]), float(row["y"]), float(row["z"])),
                }

class ReplayMonitor:
    """Replay updates from a DataSource into a DigitalTwin (offline)."""
    def __init__(self, twin: DigitalTwin, source: DataSource) -> None:
        self.twin = twin
        self.source = source

    def run(self, speed: float = 1.0) -> int:
        updates = 0
        events = sorted(self.source.read_data(), key=lambda d: d.get("timestamp_s", 0.0))
        for ev in events:
            nid = ev.get("node_id"); pos = ev.get("position")
            n = self.twin.network.get_node_by_id(nid)
            if n and pos:
                n.move_to(tuple(pos))
                updates += 1
        return updates

def rmse(a: Iterable[float], b: Iterable[float]) -> float:
    import math
    xs = list(a); ys = list(b)
    assert len(xs) == len(ys)
    return math.sqrt(sum((x - y) ** 2 for x, y in zip(xs, ys)) / len(xs))



import numpy as np
from pytwinnet.physics import FreeSpacePathLoss
from pytwinnet.accelerate.vectorized import fspl_matrix_db
from pytwinnet.core.node import WirelessNode, TransceiverProperties   # ← add TransceiverProperties
from pytwinnet.physics.environment import Environment

def test_fspl_vector_matches_scalar():
    pm = FreeSpacePathLoss()
    rng = np.random.default_rng(0)
    tx_xyz = rng.uniform(0, 300, size=(5,3))
    rx_xyz = rng.uniform(0, 300, size=(7,3))
    f = 3.5e9
    m = fspl_matrix_db(tx_xyz, rx_xyz, f)
    env = Environment(dimensions_m=(300,300,30))

    for i in range(5):
        for j in range(7):
            # Ensure the TX node uses the SAME carrier frequency as the vectorized call
            tx = WirelessNode(
                "t",
                position=tuple(tx_xyz[i]),
                transceiver_properties=TransceiverProperties(carrier_frequency_hz=f),
            )
            rx = WirelessNode("r", position=tuple(rx_xyz[j]))
            s = pm.calculate_path_loss(tx, rx, env)
            assert abs(s - m[i, j]) < 1e-5




from __future__ import annotations
from typing import Tuple, List, Optional
import numpy as np
import matplotlib.pyplot as plt
from ..core.digital_twin import DigitalTwin
from ..physics.link_budget import rx_power_dbm, noise_power_dbm, sinr_db
from ..core.node import WirelessNode

def sinr_heatmap_2d(
    twin: DigitalTwin,
    tx_id: str,
    interferer_ids: Optional[List[str]] = None,
    plane_z: float = 1.5,
    xlim: Tuple[float, float] = (0.0, 100.0),
    ylim: Tuple[float, float] = (0.0, 100.0),
    resolution: int = 100,
    bandwidth_hz: float = 20e6,
    noise_figure_db: float = 7.0,
    show: bool = False,
):
    assert twin.propagation_model is not None and twin.environment is not None
    pm = twin.propagation_model
    env = twin.environment
    tx = twin.network.get_node_by_id(tx_id)
    assert tx is not None, f"Transmitter {tx_id} not found"
    inter_nodes = [twin.network.get_node_by_id(i) for i in (interferer_ids or [])]
    inter_nodes = [n for n in inter_nodes if n is not None]

    xs = np.linspace(xlim[0], xlim[1], resolution)
    ys = np.linspace(ylim[0], ylim[1], resolution)
    X, Y = np.meshgrid(xs, ys)
    Z = np.zeros_like(X, dtype=float)

    txp = tx.transceiver_properties.transmit_power_dbm
    gt = tx.transceiver_properties.antenna_gain_dbi
    gr = 0.0
    noise_dbm = noise_power_dbm(290.0, bandwidth_hz, noise_figure_db=noise_figure_db)

    for i in range(resolution):
        for j in range(resolution):
            rx_pos = (X[i, j], Y[i, j], plane_z)
            rx = WirelessNode(node_id="__rx__", position=rx_pos)
            pl_db = pm.calculate_path_loss(tx, rx, env)
            s_dbm = rx_power_dbm(txp, gt, gr, pl_db)
            ints = []
            for ino in inter_nodes:
                pl_i = pm.calculate_path_loss(ino, rx, env)
                ints.append(rx_power_dbm(ino.transceiver_properties.transmit_power_dbm,
                                         ino.transceiver_properties.antenna_gain_dbi, gr, pl_i))
            Z[i, j] = sinr_db(s_dbm, interferers_dbm=ints, noise_dbm=noise_dbm)

    fig, ax = plt.subplots()
    im = ax.imshow(Z, origin="lower", extent=(xlim[0], xlim[1], ylim[0], ylim[1]), aspect="auto")
    ax.set_xlabel("x (m)"); ax.set_ylabel("y (m)"); ax.set_title(f"SINR Heatmap (dB) - {tx_id}")
    fig.colorbar(im, ax=ax, label="SINR (dB)")
    if show: plt.show()
    return ax, Z



from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.handover import HandoverController

def main():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(300,300,30)))
    twin.set_propagation_model(FreeSpacePathLoss())

    net = Network()
    net.add_node(WirelessNode("gNB-1", position=(50,150,10),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=32, antenna_gain_dbi=5),
                              metadata={"role":"gNB"}))
    net.add_node(WirelessNode("gNB-2", position=(250,150,10),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=32, antenna_gain_dbi=5),
                              metadata={"role":"gNB"}))
    net.add_node(WirelessNode("UE-1", position=(60,150,1.5)))
    twin.network = net

    ho = HandoverController(hysteresis_db=3.0, time_to_trigger_s=0.64)
    serving = "gNB-1"
    t = 0.0
    for step in range(11):
        x = 60 + step * 20      # move east every step
        ue = twin.network.get_node_by_id("UE-1")
        ue.move_to((x, 150, 1.5))
        serving_new = ho.step(twin, "UE-1", serving, timestamp_s=t)
        print(f"t={t:.2f}s, UE-1 at x={x}, serving={serving} -> {serving_new}")
        serving = serving_new
        t += 0.2

if __name__ == "__main__":
    main()



"""Accelerated (vectorized) utilities for PyTwinNet."""
from .vectorized import (
    fspl_matrix_db,
    rsrp_matrix_dbm,
    noise_dbm_vector,
    sinr_db_from_rsrp_matrix,
    shannon_throughput_bps_vector,
)
from .association import max_rsrp_association_vectorized

__all__ = [
    "fspl_matrix_db",
    "rsrp_matrix_dbm",
    "noise_dbm_vector",
    "sinr_db_from_rsrp_matrix",
    "shannon_throughput_bps_vector",
    "max_rsrp_association_vectorized",
]




from __future__ import annotations
from dataclasses import dataclass, field
from typing import List, Tuple

@dataclass
class Environment:
    dimensions_m: Tuple[float, float, float] = (100.0, 100.0, 30.0)
    obstacles: List[object] = field(default_factory=list)
    def is_within_bounds(self, position: Tuple[float, float, float]) -> bool:
        x, y, z = position
        w, d, h = self.dimensions_m
        return 0.0 <= x <= w and 0.0 <= y <= d and 0.0 <= z <= h



from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.scheduling import max_rsrp_association, proportional_fair_allocation

def main():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(300, 300, 30)))
    twin.set_propagation_model(FreeSpacePathLoss())

    net = Network()
    net.add_node(WirelessNode("gNB-1", position=(50, 50, 10),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=32, antenna_gain_dbi=5),
                              metadata={"role": "gNB"}))
    net.add_node(WirelessNode("gNB-2", position=(250, 250, 10),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=32, antenna_gain_dbi=5),
                              metadata={"role": "gNB"}))
    for i, pos in enumerate([(60,60,1.5),(80,80,1.5),(120,120,1.5),(200,210,1.5),(240,240,1.5),(260,260,1.5)]):
        net.add_node(WirelessNode(f"UE-{i+1}", position=pos))
    twin.network = net

    tx_ids = ["gNB-1", "gNB-2"]
    ue_ids = [f"UE-{i+1}" for i in range(6)]
    assoc = max_rsrp_association(twin, tx_ids, ue_ids)
    print("Association (UE -> TX):", assoc)

    sched = proportional_fair_allocation(twin, assoc, rb_count=20)
    for tx, ues in sched.items():
        print(f"{tx}: scheduled RBs ->", ues[:10], "...")

if __name__ == "__main__":
    main()




from __future__ import annotations
from dataclasses import dataclass
from typing import Iterable, Optional, Callable
from ..core.digital_twin import DigitalTwin
from .base import DataSource
from .mock import MockNodeUpdate

@dataclass
class RealTimeMonitor:
    twin: DigitalTwin
    source: DataSource
    on_applied: Optional[Callable[[object], None]] = None
    def poll_once(self, updates: Iterable[object] | None = None) -> int:
        count = 0
        if updates is None:
            updates = self.source.read_data()
        for upd in updates:
            applied = self._apply_update(upd)
            if applied:
                count += 1
                if self.on_applied:
                    self.on_applied(upd)
        return count
    def _apply_update(self, update: object) -> bool:
        if isinstance(update, MockNodeUpdate):
            node = self.twin.network.get_node_by_id(update.node_id)
            if node:
                x, y, z = node.position
                dx, dy, dz = update.new_position
                node.move_to((x + dx, y + dy, z + dz))
                return True
            return False
        return False



"""Basic example demonstrating PyTwinNet usage.

Steps
-----
1. Build a DigitalTwin with a simple Network (3 nodes).
2. Define a Scenario that moves one node.
3. Run the Simulator and print final node positions.
"""

from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.simulation import Scenario, MoveNodeEvent, Simulator


def main() -> None:
    # 1) Create twin and environment
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(200.0, 200.0, 30.0)))
    twin.set_propagation_model(FreeSpacePathLoss())

    # 2) Build a simple network
    net = Network()
    net.add_node(WirelessNode(node_id="gNB-1", position=(0.0, 0.0, 10.0),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=30.0)))
    net.add_node(WirelessNode(node_id="UE-1", position=(10.0, 0.0, 1.5)))
    net.add_node(WirelessNode(node_id="UE-2", position=(0.0, 20.0, 1.5)))
    twin.network = net

    # 3) Define a scenario: move UE-1
    scenario = Scenario(duration_s=10.0)
    scenario.add_event(MoveNodeEvent(timestamp=3.0, node_id="UE-1", new_position=(25.0, 5.0, 1.5)))

    # 4) Run simulator (on a snapshot to protect live twin)
    sim = Simulator(twin)
    final_twin = sim.run(scenario, copy_twin=True)

    # 5) Print final positions
    for node in final_twin.network.list_nodes():
        print(f"{node.node_id}: position={node.position}, tx_power={node.transceiver_properties.transmit_power_dbm:.1f} dBm")


if __name__ == "__main__":
    main()




from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Iterable

class DataSource(ABC):
    @abstractmethod
    def connect(self) -> None: ...
    @abstractmethod
    def read_data(self) -> Iterable[object]:
        raise NotImplementedError



from __future__ import annotations
from dataclasses import dataclass
from typing import Dict, List, Any
import random
from ..core.digital_twin import DigitalTwin
from .objective import Objective

@dataclass
class ChannelAllocationGA:
    """Toy GA for assigning channels (ints) to tx nodes to maximize objective."""
    channels: List[int]
    population: int = 30
    generations: int = 20
    mutation_p: float = 0.1
    seed: int = 0

    def optimize(self, twin: DigitalTwin, node_ids: List[str], objective: Objective) -> Dict[str, Any]:
        rng = random.Random(self.seed)

        def random_ind():
            return {nid: rng.choice(self.channels) for nid in node_ids}

        def fitness(ind):
            sim = twin.snapshot()
            for nid, ch in ind.items():
                n = sim.network.get_node_by_id(nid)
                if n:
                    n.metadata["channel"] = ch
            return objective.evaluate(sim)

        pop = [random_ind() for _ in range(self.population)]
        scores = [fitness(ind) for ind in pop]
        for _ in range(self.generations):
            parents = []
            for _ in range(self.population):
                cand = rng.sample(range(self.population), k=min(3, self.population))
                best_i = max(cand, key=lambda i: scores[i])
                parents.append(pop[best_i])
            children = []
            for i in range(0, self.population, 2):
                p1 = parents[i]
                p2 = parents[min(i + 1, self.population - 1)]
                keys = list(p1.keys())
                cut = rng.randrange(1, len(keys)) if len(keys) > 1 else 1
                c = {k: (p1[k] if j < cut else p2[k]) for j, k in enumerate(keys)}
                children.append(c)
            for c in children:
                if rng.random() < self.mutation_p:
                    k = rng.choice(list(c.keys()))
                    c[k] = rng.choice(self.channels)
            pop = children
            scores = [fitness(ind) for ind in pop]
        best_idx = max(range(self.population), key=lambda i: scores[i])
        return {"best_allocation": pop[best_idx], "best_score": scores[best_idx]}




from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Dict, Tuple, List
from ..core.digital_twin import DigitalTwin
from ..core.node import WirelessNode
from .objective import Objective
from .optimizer import Optimizer

@dataclass
class SimpleGreedyOptimizer(Optimizer):
    step_db: float = 1.0
    max_power_dbm: float = 30.0
    iterations: int = 10
    def optimize(self, twin: DigitalTwin, objective: Objective) -> Dict[str, Any]:
        best_score = objective.evaluate(twin)
        history: List[Tuple[str, float]] = []
        for _ in range(self.iterations):
            candidates = list(twin.network)
            if not candidates: break
            def proxy(n: WirelessNode) -> float:
                return float(n.metadata.get("received_traffic_mbps", 0.0))
            worst = min(candidates, key=proxy)
            props = worst.transceiver_properties
            new_pw = min(props.transmit_power_dbm + self.step_db, self.max_power_dbm)
            props.transmit_power_dbm = new_pw
            score = objective.evaluate(twin)
            history.append((worst.node_id, score))
            if score > best_score:
                best_score = score
        return {"best_score": best_score, "iterations": len(history), "metadata": {"history": history}}




from .node import WirelessNode, TransceiverProperties
from .network import Network
from .digital_twin import DigitalTwin
from .utils import list_gnbs
__all__ = ["WirelessNode","TransceiverProperties","Network","DigitalTwin","list_gnbs"]




from .environment import Environment
from .propagation import PropagationModel, FreeSpacePathLoss
from .ris import RISPanel, RISAugmentedModel
from .fading import ShadowedModel, FadedModel
from .ris_beam import SmartRISPanel, RISBeamModel
from .link_budget import rx_power_dbm, noise_power_dbm, sinr_db, shannon_throughput_bps
__all__ = ["Environment","PropagationModel","FreeSpacePathLoss",
           "rx_power_dbm", "RISPanel", "RISAugmentedModel",
             "noise_power_dbm","sinr_db","shannon_throughput_bps",
             "ShadowedModel", "FadedModel", "SmartRISPanel", "RISBeamModel"]



from __future__ import annotations
import argparse, json
from typing import Any, Dict
try:
    import yaml  # type: ignore
except Exception as e:  # pragma: no cover
    raise SystemExit("PyYAML is required. Install with: pip install pyyaml") from e

from .. import DigitalTwin, Network, WirelessNode, TransceiverProperties
from ..physics import Environment, FreeSpacePathLoss
from ..physics.fading import ShadowedModel, FadedModel
from ..physics.ris_beam import SmartRISPanel, RISBeamModel
from ..visualization.heatmap_fast import sinr_heatmap_2d_fast
from ..optimization.objective import SumThroughputObjective
from ..optimization.placement import PlacementRandomSearchOptimizer, PlacementGridOptimizer

MODEL_MAP = {
    "fspl": FreeSpacePathLoss,
    "shadowed": ShadowedModel,
    "faded": FadedModel,
    "ris_beam": RISBeamModel,
}

OBJ_MAP = {
    "sum_throughput": SumThroughputObjective,
}

def build_twin(cfg: Dict[str, Any]) -> DigitalTwin:
    twin = DigitalTwin()
    env = cfg.get("environment", {})
    dims = env.get("dimensions_m", [100.0, 100.0, 30.0])
    twin.set_environment(Environment(dimensions_m=tuple(dims)))

    pm_cfg = cfg.get("propagation", {"model": "fspl"})
    model = pm_cfg.get("model", "fspl").lower()
    if model == "fspl":
        pm = FreeSpacePathLoss()
    elif model == "shadowed":
        base = FreeSpacePathLoss()
        pm = ShadowedModel(base, sigma_db=float(pm_cfg.get("sigma_db", 6.0)))
    elif model == "faded":
        base = FreeSpacePathLoss()
        pm = FadedModel(base, kind=str(pm_cfg.get("kind", "rayleigh")), K_dB=float(pm_cfg.get("K_dB", 6.0)))
    elif model == "ris_beam":
        base = FreeSpacePathLoss()
        ris_cfg = pm_cfg.get("ris", {})
        ris = SmartRISPanel(position=tuple(ris_cfg.get("position", [50.0, 50.0, 10.0])),
                            element_count=int(ris_cfg.get("element_count", 64)))
        pm = RISBeamModel(base, ris, extra_loss_db=float(pm_cfg.get("extra_loss_db", 3.0)))
    else:
        raise ValueError(f"Unknown propagation model: {model}")
    twin.set_propagation_model(pm)

    net = Network()
    for nd in cfg.get("nodes", []):
        role = nd.get("role", "")
        node = WirelessNode(
            node_id=str(nd["id"]),
            position=tuple(nd.get("position", [0.0, 0.0, 1.5])),
            transceiver_properties=TransceiverProperties(
                transmit_power_dbm=float(nd.get("tx_power_dbm", 20.0)),
                antenna_gain_dbi=float(nd.get("ant_gain_dbi", 0.0)),
                carrier_frequency_hz=float(nd.get("f_hz", 3.5e9)),
            ),
            metadata={"role": role} if role else {},
        )
        net.add_node(node)
    twin.network = net
    return twin

def apply_scenario(twin: DigitalTwin, cfg: Dict[str, Any]) -> None:
    # Minimal scenario support
    for ev in cfg.get("scenario", {}).get("events", []):
        et = ev.get("type")
        if et == "move_node":
            nid = ev["id"]; pos = tuple(ev["position"])
            n = twin.network.get_node_by_id(nid)
            if n: n.move_to(pos)
        elif et == "set_power":
            nid = ev["id"]; p = float(ev["tx_power_dbm"])
            n = twin.network.get_node_by_id(nid)
            if n: n.transceiver_properties.transmit_power_dbm = p
        elif et == "ris_beam":
            target = ev.get("target", None)
            pm = twin.propagation_model
            if hasattr(pm, "set_beam"):
                pm.set_beam(target)  # type: ignore

def run_optimizer(twin: DigitalTwin, cfg: Dict[str, Any]) -> Dict[str, Any]:
    opt_cfg = cfg.get("optimize")
    if not opt_cfg:
        return {}
    obj_cfg = opt_cfg.get("objective", {"type": "sum_throughput", "tx_id": None})
    obj_type = obj_cfg.get("type", "sum_throughput")
    if obj_type not in OBJ_MAP:
        raise ValueError(f"Unknown objective: {obj_type}")
    obj = OBJ_MAP[obj_type](**{k: v for k, v in obj_cfg.items() if k != "type"})

    method = opt_cfg.get("method", "random_search")
    targets = opt_cfg.get("node_ids", [])
    if method == "random_search":
        bounds = opt_cfg.get("bounds", [[0, 100], [0, 100]])
        samples = int(opt_cfg.get("samples", 100))
        opt = PlacementRandomSearchOptimizer(
            bounds=((bounds[0][0], bounds[0][1]), (bounds[1][0], bounds[1][1])),
            samples=samples,
            fixed_z=float(opt_cfg.get("fixed_z", 10.0)),
        )
        return opt.optimize(twin, obj, node_ids=targets)
    elif method == "grid":
        grid_x = opt_cfg.get("grid_x", [10, 20, 30])
        grid_y = opt_cfg.get("grid_y", [10, 20, 30])
        opt = PlacementGridOptimizer(grid_x=grid_x, grid_y=grid_y, fixed_z=float(opt_cfg.get("fixed_z", 10.0)))
        return opt.optimize(twin, obj, node_ids=targets)
    else:
        raise ValueError(f"Unknown optimize.method: {method}")

def maybe_heatmap(twin: DigitalTwin, cfg: Dict[str, Any]) -> str | None:
    vis = cfg.get("visualization", {})
    if not vis.get("heatmap", {}).get("enabled", False):
        return None
    h = vis["heatmap"]
    tx = h.get("tx_id")
    interfs = h.get("interferer_ids", [])
    xlim = tuple(h.get("xlim", [0, 100]))
    ylim = tuple(h.get("ylim", [0, 100]))
    res = int(h.get("resolution", 200))
    _, _ = sinr_heatmap_2d_fast(twin, tx_id=tx, interferer_ids=interfs, xlim=xlim, ylim=ylim, resolution=res, show=False)
    out = h.get("output", "heatmap.png")
    import matplotlib.pyplot as plt
    plt.savefig(out, bbox_inches="tight", dpi=150)
    return out

def main(argv: list[str] | None = None) -> None:
    parser = argparse.ArgumentParser(prog="pytwinnet.cli")
    sub = parser.add_subparsers(dest="cmd", required=True)
    p_run = sub.add_parser("run", help="Run config-driven experiment")
    p_run.add_argument("config", help="Path to YAML config")

    args = parser.parse_args(argv)
    if args.cmd == "run":
        with open(args.config, "r", encoding="utf-8") as f:
            cfg = yaml.safe_load(f)
        twin = build_twin(cfg)
        apply_scenario(twin, cfg)
        opt_res = run_optimizer(twin, cfg)
        heat = maybe_heatmap(twin, cfg)
        report = {"optimize": opt_res, "heatmap": heat}
        out_json = cfg.get("output", "report.json")
        with open(out_json, "w", encoding="utf-8") as f:
            json.dump(report, f, indent=2)
        print(f"Saved report -> {out_json}")
        if heat:
            print(f"Saved heatmap -> {heat}")




from __future__ import annotations
from typing import Optional
import matplotlib.pyplot as plt
from ..core.network import Network

def plot_topology(network: Network, ax: Optional[plt.Axes] = None) -> plt.Axes:
    if ax is None:
        fig, ax = plt.subplots()
    for node in network:
        x, y, z = node.position
        role = str(node.metadata.get("role", "")).lower()
        if "gnb" in role or "bs" in role:
            ax.scatter([x], [y], marker="^", s=80)
        else:
            ax.scatter([x], [y], marker="o", s=50)
        ax.text(x, y, node.node_id, fontsize=8)
    ax.set_xlabel("x (m)"); ax.set_ylabel("y (m)"); ax.set_title("PyTwinNet Topology")
    return ax



from .__about__ import __version__
from .core.digital_twin import DigitalTwin
from .core.network import Network
from .core.node import WirelessNode, TransceiverProperties
__all__ = ["DigitalTwin","Network","WirelessNode","TransceiverProperties"]





from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.physics.ris import RISPanel, RISAugmentedModel
from pytwinnet.physics.link_budget import rx_power_dbm

def main():
    base = FreeSpacePathLoss()
    ris = RISPanel(position=(100, 50, 10), gain_db=12.0)
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(200, 100, 30)))
    twin.set_propagation_model(RISAugmentedModel(base, ris, extra_loss_db=3.0))

    net = Network()
    net.add_node(WirelessNode("gNB-1", position=(20, 50, 10),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=30, antenna_gain_dbi=5),
                              metadata={"role":"gNB"}))
    net.add_node(WirelessNode("UE-1", position=(180, 50, 1.5)))
    twin.network = net

    pm = twin.propagation_model; env = twin.environment
    tx = twin.network.get_node_by_id("gNB-1")
    rx = twin.network.get_node_by_id("UE-1")

    pl_ris = pm.calculate_path_loss(tx, rx, env)
    prx_ris = rx_power_dbm(tx.transceiver_properties.transmit_power_dbm,
                           tx.transceiver_properties.antenna_gain_dbi,
                           rx.transceiver_properties.antenna_gain_dbi,
                           pl_ris)

    direct_pm = FreeSpacePathLoss()
    pl_direct = direct_pm.calculate_path_loss(tx, rx, env)
    prx_direct = rx_power_dbm(tx.transceiver_properties.transmit_power_dbm,
                              tx.transceiver_properties.antenna_gain_dbi,
                              rx.transceiver_properties.antenna_gain_dbi,
                              pl_direct)

    print(f"Direct-only RSRP: {prx_direct:.2f} dBm | RIS-augmented RSRP: {prx_ris:.2f} dBm (higher is better)")

if __name__ == "__main__":
    main()



from __future__ import annotations
from dataclasses import dataclass
from typing import Callable, Dict, Any, List, Tuple

@dataclass
class MultiObjectiveSampler:
    """Sample candidates, evaluate 2 objectives, and return Pareto set (maximize both)."""
    sampler: Callable[[], Dict[str, Any]]
    evaluate: Callable[[Dict[str, Any]], Tuple[float, float]]
    samples: int = 200

    def run(self) -> Dict[str, Any]:
        pts: List[Dict[str, Any]] = []
        vals: List[Tuple[float, float]] = []
        for _ in range(self.samples):
            x = self.sampler()
            v = self.evaluate(x)
            pts.append(x); vals.append(v)
        pareto = []
        for i, v in enumerate(vals):
            dominated = False
            for j, u in enumerate(vals):
                if j == i: continue
                if u[0] >= v[0] and u[1] >= v[1] and (u[0] > v[0] or u[1] > v[1]):
                    dominated = True; break
            if not dominated:
                pareto.append((pts[i], v))
        return {"pareto": pareto, "all": list(zip(pts, vals))}



from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.handover import HandoverController

def test_ho_kpis():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(300,300,30)))
    twin.set_propagation_model(FreeSpacePathLoss())
    net = Network()
    net.add_node(WirelessNode("gNB-1",(50,150,10),TransceiverProperties(32,5),metadata={"role":"gNB"}))
    net.add_node(WirelessNode("gNB-2",(250,150,10),TransceiverProperties(32,5),metadata={"role":"gNB"}))
    net.add_node(WirelessNode("UE-1",(60,150,1.5)))
    twin.network = net

    ho = HandoverController(hysteresis_db=0.0, time_to_trigger_s=0.0)
    ho.reset_logs()

    serving = "gNB-1"
    # initial decision
    serving = ho.step_logged(twin, "UE-1", serving, 0.0)

    # move clearly toward gNB-2 and allow time to pass
    ue = twin.network.get_node_by_id("UE-1")
    ue.move_to((280,150,1.5))                     # ← further than 260 to make the margin obvious
    serving = ho.step_logged(twin, "UE-1", serving, 0.2)

    # (Optional) one more check to ensure state settled
    serving = ho.step_logged(twin, "UE-1", serving, 0.3)

    k = ho.kpis()
    assert k["handover_count"] >= 1




from pytwinnet import DigitalTwin, Network, WirelessNode
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.simulation import Scenario, MoveNodeEvent, what_if

def test_what_if_snapshot_isolated():
    twin = DigitalTwin()
    twin.set_environment(Environment())
    twin.set_propagation_model(FreeSpacePathLoss())
    net = Network()
    net.add_node(WirelessNode("UE-1", position=(0,0,0)))
    twin.network = net
    scen = Scenario(duration_s=1.0)
    scen.add_event(MoveNodeEvent(timestamp=0.5, node_id="UE-1", new_position=(10,0,0)))
    res = what_if(twin, scen)
    assert twin.network.get_node_by_id("UE-1").position == (0,0,0)
    assert res["twin"].network.get_node_by_id("UE-1").position == (10,0,0)




from __future__ import annotations
from abc import ABC, abstractmethod
from ..core.digital_twin import DigitalTwin

class Objective(ABC):
    @abstractmethod
    def evaluate(self, twin: DigitalTwin) -> float: ...

class MaximizeThroughput(Objective):
    def evaluate(self, twin: DigitalTwin) -> float:
        total = 0.0
        for node in twin.network:
            total += float(node.metadata.get("received_traffic_mbps", 0.0))
        return total

class MinimizePowerConsumption(Objective):
    def evaluate(self, twin: DigitalTwin) -> float:
        total_power = 0.0
        for node in twin.network:
            total_power += float(node.transceiver_properties.transmit_power_dbm)
        return -total_power

from ..physics.link_budget import rx_power_dbm, noise_power_dbm, sinr_db, shannon_throughput_bps
from ..physics.propagation import PropagationModel
from ..physics.environment import Environment

class SumThroughputObjective(Objective):
    def __init__(self, tx_id: str, efficiency: float = 1.0):
        self.tx_id = tx_id
        self.efficiency = efficiency
    def evaluate(self, twin: DigitalTwin) -> float:
        if twin.propagation_model is None or twin.environment is None:
            return 0.0
        tx = twin.network.get_node_by_id(self.tx_id)
        if tx is None:
            return 0.0
        pm: PropagationModel = twin.propagation_model
        env: Environment = twin.environment
        total = 0.0
        for rx in twin.network:
            if rx.node_id == self.tx_id:
                continue
            pl_db = pm.calculate_path_loss(tx, rx, env)
            txp = tx.transceiver_properties.transmit_power_dbm
            gt = tx.transceiver_properties.antenna_gain_dbi
            gr = rx.transceiver_properties.antenna_gain_dbi
            prx_dbm = rx_power_dbm(txp, gt, gr, pl_db)
            bw = float(rx.transceiver_properties.additional.get("bandwidth_hz", 20e6))
            nf = float(rx.transceiver_properties.additional.get("noise_figure_db", 7.0))
            n_dbm = noise_power_dbm(290.0, bw, nf)
            snr_db_val = sinr_db(prx_dbm, interferers_dbm=None, noise_dbm=n_dbm)
            thr = shannon_throughput_bps(bw, snr_db_val, efficiency=self.efficiency)
            total += thr
        return total




from __future__ import annotations
from dataclasses import dataclass, field
from typing import Dict, Optional, Tuple, Any

@dataclass
class TransceiverProperties:
    transmit_power_dbm: float = 20.0
    antenna_gain_dbi: float = 0.0
    carrier_frequency_hz: float = 2.4e9
    additional: Dict[str, Any] = field(default_factory=dict)

@dataclass
class WirelessNode:
    node_id: str
    position: Tuple[float, float, float] = (0.0, 0.0, 0.0)
    transceiver_properties: TransceiverProperties = field(default_factory=TransceiverProperties)
    mobility_model: Optional[object] = None
    metadata: Dict[str, Any] = field(default_factory=dict)

    def move_to(self, new_position: Tuple[float, float, float]) -> None:
        self.position = new_position

    def update_mobility(self, timestamp: float) -> None:
        if self.mobility_model and hasattr(self.mobility_model, "update"):
            self.mobility_model.update(self, timestamp)




from __future__ import annotations
import copy
from dataclasses import dataclass, field
from typing import Optional
from .network import Network
from ..physics.environment import Environment
from ..physics.propagation import PropagationModel

@dataclass
class DigitalTwin:
    network: Network = field(default_factory=Network)
    environment: Optional[Environment] = None
    propagation_model: Optional[PropagationModel] = None
    def snapshot(self) -> "DigitalTwin":
        return copy.deepcopy(self)
    def set_environment(self, env: Environment) -> None:
        self.environment = env
    def set_propagation_model(self, model: PropagationModel) -> None:
        self.propagation_model = model




from __future__ import annotations
from dataclasses import dataclass
from itertools import product
from typing import Dict, Iterable, Any, List, Tuple
from ..core.digital_twin import DigitalTwin
from .objective import Objective
from .optimizer import Optimizer

@dataclass
class GridSearchOptimizer(Optimizer):
    param_grid: Dict[str, Iterable[float]]
    copy_twin: bool = True
    def optimize(self, twin: DigitalTwin, objective: Objective) -> Dict[str, Any]:
        node_ids = list(self.param_grid.keys())
        candidates = list(product(*[list(self.param_grid[n]) for n in node_ids]))
        best_score = float("-inf")
        best_combo: List[Tuple[str, float]] = []
        for combo in candidates:
            sim_twin = twin.snapshot() if self.copy_twin else twin
            for nid, pw in zip(node_ids, combo):
                node = sim_twin.network.get_node_by_id(nid)
                if node:
                    node.transceiver_properties.transmit_power_dbm = float(pw)
            score = objective.evaluate(sim_twin)
            if score > best_score:
                best_score = score
                best_combo = list(zip(node_ids, map(float, combo)))
        return {"best_score": best_score, "best_params": best_combo, "evaluations": len(candidates)}



from __future__ import annotations
from typing import Dict, List
import numpy as np
from ..core.digital_twin import DigitalTwin
from ..accelerate.vectorized import fspl_matrix_db, rsrp_matrix_dbm, noise_dbm_vector, sinr_db_from_rsrp_matrix

def max_sinr_association(twin: DigitalTwin, tx_ids: List[str], ue_ids: List[str],
                         bandwidth_hz: float = 20e6, nf_db: float = 7.0,
                         bias_db: Dict[str, float] | None = None) -> Dict[str, str]:
    txs = [twin.network.get_node_by_id(t) for t in tx_ids]; txs = [t for t in txs if t]
    ues = [twin.network.get_node_by_id(u) for u in ue_ids]; ues = [u for u in ues if u]
    if not txs or not ues:
        return {}
    tx_xyz = np.array([t.position for t in txs], float)
    ue_xyz = np.array([u.position for u in ues], float)
    f = txs[0].transceiver_properties.carrier_frequency_hz
    pl = fspl_matrix_db(tx_xyz, ue_xyz, f)
    rsrp = rsrp_matrix_dbm(
        np.array([t.transceiver_properties.transmit_power_dbm for t in txs]),
        np.array([t.transceiver_properties.antenna_gain_dbi for t in txs]),
        np.array([u.transceiver_properties.antenna_gain_dbi for u in ues]),
        pl,
    )
    if bias_db:
        for i, t in enumerate(txs):
            rsrp[i, :] += float(bias_db.get(t.node_id, 0.0))

    noise = noise_dbm_vector(bandwidth_hz, noise_figure_db=nf_db) * np.ones(len(ues))
    best: Dict[str, str] = {}
    for j in range(len(ues)):
        sinr_per_tx = []
        for i in range(len(txs)):
            sidx = np.array([i], dtype=int)
            val = sinr_db_from_rsrp_matrix(rsrp[:, j][:, None], sidx, noise[j:j+1])[0]
            sinr_per_tx.append(val)
        best_tx = int(np.argmax(sinr_per_tx))
        best[ues[j].node_id] = txs[best_tx].node_id
    return best



from __future__ import annotations
from dataclasses import dataclass
from typing import Dict, Tuple, List, Any, Iterable
from concurrent.futures import ProcessPoolExecutor, as_completed
import itertools
import math

from ..core.digital_twin import DigitalTwin
from .objective import Objective

Pos = Tuple[float, float, float]

def _eval_combo(snapshot: DigitalTwin, node_ids: List[str], coords: List[Pos], objective: Objective) -> Tuple[float, Dict[str, Pos]]:
    sim = snapshot  # already deep-copied by caller
    for nid, pos in zip(node_ids, coords):
        n = sim.network.get_node_by_id(nid)
        if n: n.move_to(pos)
    score = objective.evaluate(sim)
    return score, {nid: pos for nid, pos in zip(node_ids, coords)}

@dataclass
class ParallelPlacementGrid:
    grid_x: Iterable[float]
    grid_y: Iterable[float]
    fixed_z: float = 10.0
    max_workers: int = 0  # 0 -> use os.cpu_count()

    def optimize(self, twin: DigitalTwin, objective: Objective, node_ids: List[str]) -> Dict[str, Any]:
        points = [(x, y, self.fixed_z) for x, y in itertools.product(self.grid_x, self.grid_y)]
        combos = itertools.product(points, repeat=len(node_ids))
        best_score = -math.inf
        best_pos: Dict[str, Pos] = {}
        futures = []
        with ProcessPoolExecutor(max_workers=self.max_workers or None) as ex:
            for combo in combos:
                # Snapshot per job (so workers don't share state)
                snap = twin.snapshot()
                futures.append(ex.submit(_eval_combo, snap, node_ids, list(combo), objective))
            for fut in as_completed(futures):
                score, pos = fut.result()
                if score > best_score:
                    best_score, best_pos = score, pos
        return {"best_score": best_score, "best_positions": best_pos, "evaluations": len(futures)}



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Callable, List, Tuple, Any
import numpy as np
import math
from typing import Dict

def rbf_kernel(X1: np.ndarray, X2: np.ndarray, lengthscale: float, variance: float) -> np.ndarray:
    D = ((X1[:, None, :] - X2[None, :, :]) ** 2).sum(axis=2)
    return variance * np.exp(-0.5 * D / (lengthscale ** 2 + 1e-12))

@dataclass
class SimpleBayesOpt:
    """Tiny Bayesian Optimization (RBF GP + Expected Improvement)."""
    bounds: List[Tuple[float, float]]
    init_points: int = 8
    iters: int = 32
    lengthscale: float = 0.5
    variance: float = 1.0
    noise: float = 1e-6
    seed: int = 0

    X: List[List[float]] = field(default_factory=list)
    y: List[float] = field(default_factory=list)

    def ask(self, n: int = 1) -> List[List[float]]:
        rng = np.random.default_rng(self.seed + len(self.X))
        pts = []
        for _ in range(n):
            x = [rng.uniform(a, b) for (a, b) in self.bounds]
            pts.append(x)
        return pts

    def tell(self, X: List[List[float]], y: List[float]) -> None:
        self.X.extend(X); self.y.extend(y)

    def _posterior(self, Xs: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        if not self.X:
            mu = np.zeros((Xs.shape[0],)); s2 = np.ones_like(mu)
            return mu, s2
        X = np.array(self.X, float); y = np.array(self.y, float)
        K = rbf_kernel(X, X, self.lengthscale, self.variance) + self.noise * np.eye(X.shape[0])
        Ks = rbf_kernel(X, Xs, self.lengthscale, self.variance)
        Kss = rbf_kernel(Xs, Xs, self.lengthscale, self.variance) + self.noise * np.eye(Xs.shape[0])
        L = np.linalg.cholesky(K)
        alpha = np.linalg.solve(L.T, np.linalg.solve(L, y))
        mu = Ks.T @ alpha
        v = np.linalg.solve(L, Ks)
        s2 = np.maximum(np.diag(Kss) - (v * v).sum(axis=0), 1e-12)
        return mu, s2

    def _ei(self, mu: np.ndarray, s2: np.ndarray, y_best: float) -> np.ndarray:
            # Vectorized Expected Improvement
            s = np.sqrt(s2)
            z = (mu - y_best) / (s + 1e-12)
            erf = np.vectorize(math.erf)
            Phi = 0.5 * (1.0 + erf(z / np.sqrt(2.0)))                  # CDF of N(0,1)
            phi = (1.0 / np.sqrt(2.0 * np.pi)) * np.exp(-0.5 * z * z)  # PDF of N(0,1)
            return (mu - y_best) * Phi + s * phi


    def suggest(self, n: int = 1) -> List[List[float]]:
        rng = np.random.default_rng(self.seed + 1234 + len(self.X))
        cand = rng.uniform(
            low=np.array([a for a, _ in self.bounds]),
            high=np.array([b for _, b in self.bounds]),
            size=(1024, len(self.bounds)),
        )
        mu, s2 = self._posterior(cand)
        y_best = max(self.y) if self.y else 0.0
        ei = self._ei(mu, s2, y_best)
        idx = np.argsort(-ei)[:n]
        return cand[idx].tolist()

    def run(self, evaluate: Callable[[List[float]], float]) -> Dict[str, Any]:
        X0 = self.ask(self.init_points)
        y0 = [evaluate(x) for x in X0]
        self.tell(X0, y0)
        for _ in range(self.iters):
            Xn = self.suggest(1)
            yn = [evaluate(x) for x in Xn]
            self.tell(Xn, yn)
        best_i = int(np.argmax(self.y))
        return {"best_x": self.X[best_i], "best_y": float(self.y[best_i]), "n_evals": len(self.y)}



from __future__ import annotations
from typing import List

def round_robin_schedule(ue_ids: List[str], rb_count: int) -> List[str]:
    if not ue_ids:
        return []
    out: List[str] = []
    i = 0
    for _ in range(rb_count):
        out.append(ue_ids[i])
        i = (i + 1) % len(ue_ids)
    return out




from .events import Event, MoveNodeEvent, TrafficGenerationEvent
from .scenario import Scenario
from .simulator import Simulator
from .what_if import what_if
__all__ = ["Event","MoveNodeEvent","TrafficGenerationEvent","Scenario","Simulator","what_if"]



from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.optimization.objective import SumThroughputObjective
from pytwinnet.optimization.bayesopt import SimpleBayesOpt

def main():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(300,300,30)))
    twin.set_propagation_model(FreeSpacePathLoss())
    net = Network()
    net.add_node(WirelessNode("gNB-1",(50,50,10),TransceiverProperties(32,5), metadata={"role":"gNB"}))
    for i, pos in enumerate([(60,80,1.5),(120,90,1.5),(220,150,1.5),(260,240,1.5)]):
        net.add_node(WirelessNode(f"UE-{i+1}", position=pos))
    twin.network = net

    obj = SumThroughputObjective(tx_id="gNB-1", efficiency=0.75)

    def evaluate_xy(xy):
        x, y = xy
        sim = twin.snapshot()
        sim.network.get_node_by_id("gNB-1").move_to((float(x), float(y), 10.0))
        return obj.evaluate(sim)

    bo = SimpleBayesOpt(bounds=[(0,300),(0,300)], init_points=8, iters=24, lengthscale=80.0, variance=1.0, seed=0)
    res = bo.run(evaluate_xy)
    print("BO best:", res)

if __name__ == "__main__":
    main()




from __future__ import annotations
from dataclasses import dataclass
from typing import Dict, Tuple, Any, List
import random
from ..core.digital_twin import DigitalTwin
from .objective import Objective
from .optimizer import Optimizer

@dataclass
class RandomSearchOptimizer(Optimizer):
    ranges_dbm: Dict[str, Tuple[float, float]]
    samples: int = 32
    seed: int = 0
    copy_twin: bool = True
    def optimize(self, twin: DigitalTwin, objective: Objective) -> Dict[str, Any]:
        rng = random.Random(self.seed)
        node_ids = list(self.ranges_dbm.keys())
        best_score = float("-inf")
        best_params: List[Tuple[str, float]] = []
        for _ in range(self.samples):
            sim_twin = twin.snapshot() if self.copy_twin else twin
            draw = []
            for nid in node_ids:
                mn, mx = self.ranges_dbm[nid]
                val = rng.uniform(float(mn), float(mx))
                node = sim_twin.network.get_node_by_id(nid)
                if node:
                    node.transceiver_properties.transmit_power_dbm = float(val)
                draw.append((nid, float(val)))
            score = objective.evaluate(sim_twin)
            if score > best_score:
                best_score = score
                best_params = draw
        return {"best_score": best_score, "best_params": best_params, "evaluations": self.samples}




from __future__ import annotations
import math
from abc import ABC, abstractmethod
from ..core.node import WirelessNode
from .environment import Environment

class PropagationModel(ABC):
    @abstractmethod
    def calculate_path_loss(self, tx: WirelessNode, rx: WirelessNode, environment: Environment) -> float: ...

class FreeSpacePathLoss(PropagationModel):
    def calculate_path_loss(self, tx: WirelessNode, rx: WirelessNode, environment: Environment) -> float:
        dx = tx.position[0] - rx.position[0]
        dy = tx.position[1] - rx.position[1]
        dz = tx.position[2] - rx.position[2]
        d_m = math.sqrt(dx*dx + dy*dy + dz*dz)
        d_km = max(d_m, 1e-3) / 1000.0
        f_mhz = max(tx.transceiver_properties.carrier_frequency_hz, 1.0) / 1e6
        return 20.0 * math.log10(d_km) + 20.0 * math.log10(f_mhz) + 32.44




from .objective import Objective, MaximizeThroughput, MinimizePowerConsumption, SumThroughputObjective
from .optimizer import Optimizer
from .simple_greedy import SimpleGreedyOptimizer
from .grid_search import GridSearchOptimizer
from .random_search import RandomSearchOptimizer
__all__ = ["Objective","MaximizeThroughput","MinimizePowerConsumption","SumThroughputObjective",
           "Optimizer","SimpleGreedyOptimizer","GridSearchOptimizer","RandomSearchOptimizer"]




from pytwinnet.mobility import ConstantVelocity
from pytwinnet.core.node import WirelessNode

def test_constant_velocity_moves_node():
    n = WirelessNode("ue", position=(0,0,0))
    m = ConstantVelocity(velocity_mps=(1.0, 0.0, 0.0))
    n.mobility_model = m
    n.update_mobility(0.0)   # init
    n.update_mobility(2.0)   # after 2s
    assert abs(n.position[0] - 2.0) < 1e-6



from __future__ import annotations
from typing import List, Iterable
import numpy as np
from ..core.node import WirelessNode
from ..core.digital_twin import DigitalTwin
from .propagation import PropagationModel, FreeSpacePathLoss
from ..accelerate.vectorized import fspl_matrix_db

def path_loss_matrix(
    twin: DigitalTwin,
    tx_nodes: List[WirelessNode],
    rx_positions: np.ndarray,  # (R,3) float
) -> np.ndarray:
    """
    Return PL (dB) matrix of shape (T,R) for arbitrary PropagationModel.

    Fast paths:
      - If model is FreeSpacePathLoss -> fully vectorized (no Python loops).
      - If model implements an optional `calculate_path_loss_batch(tx_nodes, rx_positions, env)`
        it will be used directly (hook for future custom vectorized models).

    Fallback:
      - Reuse a single RX WirelessNode and update its .position per point to avoid
        per-point allocations. This is still scalar but much faster than constructing
        nodes in a tight loop.
    """
    pm: PropagationModel = twin.propagation_model
    env = twin.environment
    assert pm is not None and env is not None

    T = len(tx_nodes)
    R = int(rx_positions.shape[0])
    if T == 0 or R == 0:
        return np.zeros((T, R), dtype=float)

    # Fast path: pure FSPL
    if isinstance(pm, FreeSpacePathLoss):
        tx_xyz = np.array([tx.position for tx in tx_nodes], dtype=float)  # (T,3)
        f_hz = tx_nodes[0].transceiver_properties.carrier_frequency_hz
        return fspl_matrix_db(tx_xyz, rx_positions, f_hz)

    # Optional vectorized API hook (future-friendly)
    if hasattr(pm, "calculate_path_loss_batch"):
        return pm.calculate_path_loss_batch(tx_nodes, rx_positions, env)  # type: ignore[attr-defined]

    # Scalar fallback (minimize allocations)
    pl = np.empty((T, R), dtype=float)
    rx = WirelessNode("__rx__", position=(0.0, 0.0, 0.0))
    for j in range(R):
        # update one shared RX node position
        rx.position = (float(rx_positions[j, 0]), float(rx_positions[j, 1]), float(rx_positions[j, 2]))
        for i, tx in enumerate(tx_nodes):
            pl[i, j] = pm.calculate_path_loss(tx, rx, env)
    return pl



from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.simulation import Scenario, MoveNodeEvent, what_if
from pytwinnet.optimization.objective import SumThroughputObjective

def build_twin():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(300.0, 300.0, 30.0)))
    twin.set_propagation_model(FreeSpacePathLoss())
    net = Network()
    net.add_node(WirelessNode(node_id="gNB-1", position=(0.0, 0.0, 10.0),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=30.0, antenna_gain_dbi=5.0)))
    for i, pos in enumerate([(30.0, 0.0, 1.5), (0.0, 40.0, 1.5), (25.0, 25.0, 1.5), (60.0, 50.0, 1.5)]):
        net.add_node(WirelessNode(node_id=f"UE-{i+1}", position=pos))
    twin.network = net
    return twin

def scenario_move_gnb(new_position):
    s = Scenario(duration_s=1.0)
    s.add_event(MoveNodeEvent(timestamp=0.5, node_id="gNB-1", new_position=new_position))
    return s

def main():
    twin = build_twin()
    obj = SumThroughputObjective(tx_id="gNB-1", efficiency=0.75)
    resA = what_if(twin, scenario_move_gnb((20.0, 20.0, 10.0)), objective=obj)
    resB = what_if(twin, scenario_move_gnb((80.0, 80.0, 10.0)), objective=obj)
    print("Candidate A score (bps):", f"{resA['score']:.2f}")
    print("Candidate B score (bps):", f"{resB['score']:.2f}")
    print("\nFinal positions for Candidate A:")
    for n in resA["twin"].network.list_nodes():
        print(f"  {n.node_id}: {n.position}")

if __name__ == "__main__":
    main()



from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.physics.fading import ShadowedModel, FadedModel
from pytwinnet.physics.link_budget import rx_power_dbm

def main():
    base = FreeSpacePathLoss()
    model = ShadowedModel(FadedModel(base, kind="rician", K_dB=6.0, seed=42), sigma_db=5.0, seed=7)
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(200,200,30)))
    twin.set_propagation_model(model)

    net = Network()
    net.add_node(WirelessNode("gNB-1", position=(0,0,10),
                 transceiver_properties=TransceiverProperties(transmit_power_dbm=30, antenna_gain_dbi=5),
                 metadata={"role":"gNB"}))
    net.add_node(WirelessNode("UE-1", position=(120,80,1.5)))
    twin.network = net

    pm = twin.propagation_model; env = twin.environment
    tx = twin.network.get_node_by_id("gNB-1"); rx = twin.network.get_node_by_id("UE-1")

    # epoch 0
    pm.set_epoch(0)
    pl0 = pm.calculate_path_loss(tx, rx, env)
    p0 = rx_power_dbm(tx.transceiver_properties.transmit_power_dbm, tx.transceiver_properties.antenna_gain_dbi, 0.0, pl0)

    # epoch 1 (refresh shadow/fading sample)
    pm.set_epoch(1)
    pl1 = pm.calculate_path_loss(tx, rx, env)
    p1 = rx_power_dbm(tx.transceiver_properties.transmit_power_dbm, tx.transceiver_properties.antenna_gain_dbi, 0.0, pl1)

    print(f"RSRP epoch0={p0:.2f} dBm, epoch1={p1:.2f} dBm (should differ due to fading/shadowing)")

if __name__ == "__main__":
    main()



from __future__ import annotations
from typing import Dict, List
from ..core.digital_twin import DigitalTwin
from ..physics.link_budget import rx_power_dbm
from ..physics.propagation import PropagationModel
from ..physics.environment import Environment

def max_rsrp_association(twin: DigitalTwin, tx_ids: List[str], ue_ids: List[str]) -> Dict[str, str]:
    """Associate each UE to the TX (gNB) with the strongest RSRP."""
    pm: PropagationModel = twin.propagation_model
    env: Environment = twin.environment
    out: Dict[str, str] = {}
    for ue_id in ue_ids:
        ue = twin.network.get_node_by_id(ue_id)
        best_tx, best_p = None, float("-inf")
        for tx_id in tx_ids:
            tx = twin.network.get_node_by_id(tx_id)
            pl_db = pm.calculate_path_loss(tx, ue, env)
            prx_dbm = rx_power_dbm(
                tx.transceiver_properties.transmit_power_dbm,
                tx.transceiver_properties.antenna_gain_dbi,
                ue.transceiver_properties.antenna_gain_dbi,
                pl_db,
            )
            if prx_dbm > best_p:
                best_p, best_tx = prx_dbm, tx_id
        if best_tx is not None:
            out[ue_id] = best_tx
    return out




from __future__ import annotations
import random, math
from dataclasses import dataclass
from typing import Tuple, Optional
from ..core.node import WirelessNode
from ..physics.environment import Environment

@dataclass
class RandomWaypoint:
    env: Environment
    speed_range_mps: Tuple[float, float] = (0.5, 1.5)
    pause_time_s: float = 0.0
    seed: int = 0
    _rng: random.Random = None
    _target: Optional[Tuple[float, float, float]] = None
    _speed: float = 0.0
    _paused_until: float = 0.0
    _last_t: Optional[float] = None
    def __post_init__(self):
        self._rng = random.Random(self.seed)
    def _sample_waypoint(self) -> Tuple[float, float, float]:
        w, d, h = self.env.dimensions_m
        return (self._rng.uniform(0, w), self._rng.uniform(0, d), self._rng.uniform(0, h))
    def update(self, node: WirelessNode, timestamp: float) -> None:
        if self._last_t is None:
            self._last_t = timestamp
            self._target = self._sample_waypoint()
            self._speed = self._rng.uniform(*self.speed_range_mps)
            return
        dt = max(0.0, timestamp - self._last_t)
        self._last_t = timestamp
        if timestamp < self._paused_until:
            return
        x, y, z = node.position
        tx, ty, tz = self._target
        dx, dy, dz = tx - x, ty - y, tz - z
        dist = math.sqrt(dx*dx + dy*dy + dz*dz)
        if dist < 1e-6:
            self._paused_until = timestamp + self.pause_time_s
            self._target = self._sample_waypoint()
            self._speed = self._rng.uniform(*self.speed_range_mps)
            return
        step = self._speed * dt
        if step >= dist:
            node.move_to((tx, ty, tz))
        else:
            nx = x + dx / dist * step
            ny = y + dy / dist * step
            nz = z + dz / dist * step
            node.move_to((nx, ny, nz))



from __future__ import annotations
from dataclasses import dataclass
from typing import Dict
import numpy as np
from .objective import Objective
from ..accelerate.vectorized import (
    fspl_matrix_db, rsrp_matrix_dbm, noise_dbm_vector, sinr_db_from_rsrp_matrix, shannon_throughput_bps_vector
)

@dataclass
class ProportionalFairObjective(Objective):
    tx_id: str
    efficiency: float = 0.75
    rb_bandwidth_hz: float = 180e3
    noise_figure_db: float = 7.0
    avg_throughput_bps: Dict[str, float] | None = None

    def evaluate(self, twin) -> float:
        tx = twin.network.get_node_by_id(self.tx_id)
        ues = [n for n in twin.network if n.node_id != self.tx_id]
        if not tx or not ues: return -1e9
        tx_xyz = np.array([tx.position], float)
        ue_xyz = np.array([u.position for u in ues], float)
        pl = fspl_matrix_db(tx_xyz, ue_xyz, tx.transceiver_properties.carrier_frequency_hz)
        rsrp = rsrp_matrix_dbm(
            np.array([tx.transceiver_properties.transmit_power_dbm]),
            np.array([tx.transceiver_properties.antenna_gain_dbi]),
            np.array([u.transceiver_properties.antenna_gain_dbi for u in ues]),
            pl,
        )[0]
        noise = noise_dbm_vector(self.rb_bandwidth_hz, noise_figure_db=self.noise_figure_db) * np.ones(len(ues))
        sinr = sinr_db_from_rsrp_matrix(rsrp[None, :], np.zeros(len(ues), dtype=int), noise)
        thr = shannon_throughput_bps_vector(self.rb_bandwidth_hz, sinr, efficiency=self.efficiency)
        thr = np.maximum(thr, 1.0)
        return float(np.sum(np.log(thr)))

@dataclass
class CoveragePercentileObjective(Objective):
    tx_id: str
    percentile: float = 5.0
    bandwidth_hz: float = 20e6
    noise_figure_db: float = 7.0

    def evaluate(self, twin) -> float:
        tx = twin.network.get_node_by_id(self.tx_id)
        ues = [n for n in twin.network if n.node_id != self.tx_id]
        if not tx or not ues: return -1e9
        tx_xyz = np.array([tx.position], float)
        ue_xyz = np.array([u.position for u in ues], float)
        pl = fspl_matrix_db(tx_xyz, ue_xyz, tx.transceiver_properties.carrier_frequency_hz)
        rsrp = rsrp_matrix_dbm(
            np.array([tx.transceiver_properties.transmit_power_dbm]),
            np.array([tx.transceiver_properties.antenna_gain_dbi]),
            np.array([u.transceiver_properties.antenna_gain_dbi for u in ues]),
            pl,
        )[0]
        noise = noise_dbm_vector(self.bandwidth_hz, noise_figure_db=self.noise_figure_db) * np.ones(len(ues))
        sinr = sinr_db_from_rsrp_matrix(rsrp[None, :], np.zeros(len(ues), dtype=int), noise)
        return float(np.percentile(sinr, self.percentile))

@dataclass
class EnergyEfficiencyObjective(Objective):
    tx_id: str
    efficiency: float = 0.75
    bandwidth_hz: float = 20e6
    noise_figure_db: float = 7.0
    power_penalty: float = 1.0  # utility penalty per dBm

    def evaluate(self, twin) -> float:
        tx = twin.network.get_node_by_id(self.tx_id)
        ues = [n for n in twin.network if n.node_id != self.tx_id]
        if not tx or not ues: return -1e9
        tx_xyz = np.array([tx.position], float)
        ue_xyz = np.array([u.position for u in ues], float)
        pl = fspl_matrix_db(tx_xyz, ue_xyz, tx.transceiver_properties.carrier_frequency_hz)
        rsrp = rsrp_matrix_dbm(
            np.array([tx.transceiver_properties.transmit_power_dbm]),
            np.array([tx.transceiver_properties.antenna_gain_dbi]),
            np.array([u.transceiver_properties.antenna_gain_dbi for u in ues]),
            pl,
        )[0]
        noise = noise_dbm_vector(self.bandwidth_hz, noise_figure_db=self.noise_figure_db) * np.ones(len(ues))
        sinr = sinr_db_from_rsrp_matrix(rsrp[None, :], np.zeros(len(ues), dtype=int), noise)
        thr = shannon_throughput_bps_vector(self.bandwidth_hz, sinr, efficiency=self.efficiency)
        utility = float(np.sum(thr))
        penalty = self.power_penalty * tx.transceiver_properties.transmit_power_dbm
        return utility - penalty




from __future__ import annotations
from typing import Dict, Any
from .simulator import Simulator
from .scenario import Scenario
from ..core.digital_twin import DigitalTwin
from ..optimization.objective import Objective

def what_if(twin: DigitalTwin, scenario: Scenario, objective: Objective | None = None) -> Dict[str, Any]:
    sim = Simulator(twin)
    out = sim.run(scenario, copy_twin=True)
    score = objective.evaluate(out) if objective is not None else None
    return {"twin": out, "score": score}



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Dict
from ..core.digital_twin import DigitalTwin
from ..physics.link_budget import rx_power_dbm

@dataclass
class HandoverController:
    hysteresis_db: float = 3.0
    time_to_trigger_s: float = 0.64
    _cand: Dict[str, Dict[str, float]] = field(default_factory=dict)  # ue_id -> {candidate_tx, start_time}

    def step(self, twin: DigitalTwin, ue_id: str, current_tx_id: str, timestamp_s: float) -> str:
        pm = twin.propagation_model; env = twin.environment
        ue = twin.network.get_node_by_id(ue_id)
        cur = twin.network.get_node_by_id(current_tx_id)

        # current RSRP
        cur_pl = pm.calculate_path_loss(cur, ue, env)
        cur_rsrp = rx_power_dbm(
            cur.transceiver_properties.transmit_power_dbm,
            cur.transceiver_properties.antenna_gain_dbi,
            ue.transceiver_properties.antenna_gain_dbi,
            cur_pl,
        )

        # best neighbor RSRP among gNB/BS roles
        best_tx_id, best_rsrp = current_tx_id, cur_rsrp
        for tx in twin.network:
            if tx.node_id == current_tx_id:
                continue
            role = str(tx.metadata.get("role", "")).lower()
            if "gnb" not in role and "bs" not in role:
                continue
            pl = pm.calculate_path_loss(tx, ue, env)
            p = rx_power_dbm(
                tx.transceiver_properties.transmit_power_dbm,
                tx.transceiver_properties.antenna_gain_dbi,
                ue.transceiver_properties.antenna_gain_dbi,
                pl,
            )
            if p > best_rsrp:
                best_tx_id, best_rsrp = tx.node_id, p

        # hysteresis + TTT
        if best_rsrp >= cur_rsrp + self.hysteresis_db and best_tx_id != current_tx_id:
            rec = self._cand.get(ue_id)
            if rec is None or rec.get("candidate_tx") != best_tx_id:
                self._cand[ue_id] = {"candidate_tx": best_tx_id, "start_time": timestamp_s}
            else:
                if (timestamp_s - rec["start_time"]) >= self.time_to_trigger_s:
                    self._cand.pop(ue_id, None)
                    return best_tx_id
        else:
            self._cand.pop(ue_id, None)

        return current_tx_id
    
        # --- KPI logging helpers ---
    def reset_logs(self):
        self._log = {"events": [], "ping_pong": 0}

    def log_event(self, ue_id: str, from_tx: str, to_tx: str, t: float):
        if not hasattr(self, "_log"):
            self.reset_logs()
        self._log["events"].append({"t": t, "ue": ue_id, "from": from_tx, "to": to_tx})
        if len(self._log["events"]) >= 2:
            prev = self._log["events"][-2]
            if prev["ue"] == ue_id and prev["from"] == to_tx and prev["to"] == from_tx:
                self._log["ping_pong"] += 1

    def step_logged(self, twin, ue_id: str, current_tx_id: str, timestamp_s: float) -> str:
        new_tx = self.step(twin, ue_id, current_tx_id, timestamp_s)
        if new_tx != current_tx_id:
            self.log_event(ue_id, current_tx_id, new_tx, timestamp_s)
        return new_tx

    def kpis(self) -> dict:
        if not hasattr(self, "_log"):
            self.reset_logs()
        return {
            "handover_count": len(self._log["events"]),
            "ping_pong_count": self._log["ping_pong"],
        }




from __future__ import annotations
from dataclasses import dataclass
from itertools import product
from typing import Dict, Iterable, List, Tuple, Any
import random
from ..core.digital_twin import DigitalTwin
from ..optimization.objective import Objective

Bounds2D = Tuple[Tuple[float,float], Tuple[float,float]]  # ((x_min,x_max),(y_min,y_max))

@dataclass
class PlacementGridOptimizer:
    """
    Grid search over positions for a set of nodes (e.g., gNBs).
    Beware combinatorics: keep grids small or optimize one node at a time.
    """
    grid_x: Iterable[float]
    grid_y: Iterable[float]
    fixed_z: float = 10.0
    copy_twin: bool = True

    def optimize(self, twin: DigitalTwin, objective: Objective, node_ids: List[str]) -> Dict[str, Any]:
        best_score = float("-inf")
        best_coords: Dict[str, Tuple[float,float,float]] = {}
        points = [(x,y,self.fixed_z) for x in self.grid_x for y in self.grid_y]
        combos = list(product(points, repeat=len(node_ids)))
        evals = 0
        for combo in combos:
            sim = twin.snapshot() if self.copy_twin else twin
            for nid, pos in zip(node_ids, combo):
                n = sim.network.get_node_by_id(nid)
                if n: n.move_to(pos)
            score = objective.evaluate(sim)
            evals += 1
            if score > best_score:
                best_score = score
                best_coords = {nid: pos for nid, pos in zip(node_ids, combo)}
        return {"best_score": best_score, "best_positions": best_coords, "evaluations": evals}

@dataclass
class PlacementRandomSearchOptimizer:
    """
    Randomly sample positions within bounds for given nodes.
    """
    bounds: Bounds2D
    samples: int = 200
    fixed_z: float = 10.0
    seed: int = 0
    copy_twin: bool = True

    def optimize(self, twin: DigitalTwin, objective: Objective, node_ids: List[str]) -> Dict[str, Any]:
        rng = random.Random(self.seed)
        (x0,x1),(y0,y1) = self.bounds
        best_score = float("-inf")
        best_coords: Dict[str, Tuple[float,float,float]] = {}
        for _ in range(self.samples):
            sim = twin.snapshot() if self.copy_twin else twin
            cur_coords = {}
            for nid in node_ids:
                x = rng.uniform(x0, x1); y = rng.uniform(y0, y1); z = self.fixed_z
                n = sim.network.get_node_by_id(nid)
                if n: n.move_to((x,y,z))
                cur_coords[nid] = (x,y,z)
            score = objective.evaluate(sim)
            if score > best_score:
                best_score = score
                best_coords = cur_coords
        return {"best_score": best_score, "best_positions": best_coords, "evaluations": self.samples}



import time, pytest
from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.visualization.heatmap_fast import sinr_heatmap_2d_fast

@pytest.mark.perf
def test_heatmap_perf_smoke():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(400,400,30)))
    twin.set_propagation_model(FreeSpacePathLoss())
    net = Network()
    net.add_node(WirelessNode("gNB-1",(100,100,10),TransceiverProperties(32,5),metadata={"role":"gNB"}))
    net.add_node(WirelessNode("gNB-2",(300,300,10),TransceiverProperties(32,5),metadata={"role":"gNB"}))
    twin.network = net

    t0 = time.time()
    _, _ = sinr_heatmap_2d_fast(twin, "gNB-1", interferer_ids=["gNB-2"], resolution=250, show=False)
    dt = time.time() - t0
    assert dt < 2.0




from __future__ import annotations
from dataclasses import dataclass, field
from typing import List
from .events import Event

@dataclass
class Scenario:
    duration_s: float
    events: List[Event] = field(default_factory=list)
    def add_event(self, event: Event) -> None:
        self.events.append(event)
    def sorted_events(self) -> List[Event]:
        return sorted(self.events, key=lambda e: e.timestamp)




from __future__ import annotations
from dataclasses import dataclass
from ..core.digital_twin import DigitalTwin
from .scenario import Scenario

@dataclass
class Simulator:
    twin: DigitalTwin
    def run(self, scenario: Scenario, copy_twin: bool = True) -> DigitalTwin:
        sim_twin = self.twin.snapshot() if copy_twin else self.twin
        for event in scenario.sorted_events():
            if 0.0 <= event.timestamp <= scenario.duration_s:
                event.apply(sim_twin)
        return sim_twin



from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.visualization.heatmap_fast import sinr_heatmap_2d_fast

def main():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(400, 400, 30)))
    twin.set_propagation_model(FreeSpacePathLoss())

    net = Network()
    net.add_node(WirelessNode("gNB-1", position=(120, 150, 10),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=32, antenna_gain_dbi=5),
                              metadata={"role":"gNB"}))
    net.add_node(WirelessNode("gNB-2", position=(300, 250, 10),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=32, antenna_gain_dbi=5),
                              metadata={"role":"gNB"}))
    twin.network = net

    sinr_heatmap_2d_fast(
        twin, tx_id="gNB-1", interferer_ids=["gNB-2"],
        xlim=(0, 400), ylim=(0, 400), resolution=250, show=False
    )
    print("Computed fast heatmap at 250x250 resolution.")

if __name__ == "__main__":
    main()



from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.physics.ris_beam import SmartRISPanel, RISBeamModel
from pytwinnet.physics.link_budget import rx_power_dbm

def main():
    base = FreeSpacePathLoss()
    ris = SmartRISPanel(position=(100, 100, 10), element_count=128)
    model = RISBeamModel(base, ris, extra_loss_db=3.0)

    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(200,200,30)))
    twin.set_propagation_model(model)

    net = Network()
    net.add_node(WirelessNode("gNB-1", position=(20, 100, 10),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=30, antenna_gain_dbi=5),
                              metadata={"role":"gNB"}))
    net.add_node(WirelessNode("UE-1", position=(180, 100, 1.5)))
    net.add_node(WirelessNode("UE-2", position=(180, 130, 1.5)))
    twin.network = net

    pm = twin.propagation_model; env = twin.environment
    tx = net.get_node_by_id("gNB-1")

    # Steer RIS toward UE-1
    pm.set_beam("UE-1")
    for ue_id in ["UE-1","UE-2"]:
        rx = net.get_node_by_id(ue_id)
        pl = pm.calculate_path_loss(tx, rx, env)
        prx = rx_power_dbm(tx.transceiver_properties.transmit_power_dbm,
                           tx.transceiver_properties.antenna_gain_dbi, 0.0, pl)
        print(f"With beam=UE-1, RSRP at {ue_id}: {prx:.2f} dBm")

    # Steer RIS toward UE-2
    pm.set_beam("UE-2")
    for ue_id in ["UE-1","UE-2"]:
        rx = net.get_node_by_id(ue_id)
        pl = pm.calculate_path_loss(tx, rx, env)
        prx = rx_power_dbm(tx.transceiver_properties.transmit_power_dbm,
                           tx.transceiver_properties.antenna_gain_dbi, 0.0, pl)
        print(f"With beam=UE-2, RSRP at {ue_id}: {prx:.2f} dBm")

if __name__ == "__main__":
    main()



from __future__ import annotations
import math, hashlib, random
from dataclasses import dataclass
from typing import Optional
from .propagation import PropagationModel
from .environment import Environment
from ..core.node import WirelessNode

def _pair_seed(base_seed: int, a: str, b: str, epoch: Optional[int]) -> int:
    s = f"{base_seed}|{a}|{b}|{epoch if epoch is not None else 'static'}"
    h = hashlib.sha256(s.encode()).hexdigest()
    return int(h[:16], 16)  # 64-bit

@dataclass
class ShadowedModel(PropagationModel):
    """
    Wraps a base propagation model and adds *log-normal shadowing* (Gaussian in dB).
    Shadowing is deterministic per (tx, rx, epoch) for reproducibility. Change 'epoch'
    (int) to refresh the samples (e.g., per-drop or per-time-slot).
    """
    base: PropagationModel
    sigma_db: float = 6.0
    seed: int = 0
    epoch: Optional[int] = None

    def set_epoch(self, epoch: Optional[int]) -> None:
        self.epoch = epoch

    def calculate_path_loss(self, tx: WirelessNode, rx: WirelessNode, environment: Environment) -> float:
        pl = self.base.calculate_path_loss(tx, rx, environment)
        rng = random.Random(_pair_seed(self.seed, tx.node_id, rx.node_id, self.epoch))
        shad = rng.gauss(0.0, self.sigma_db)
        return pl + shad

@dataclass
class FadedModel(PropagationModel):
    """
    Wraps a base model and adds *small-scale fading* (Rayleigh or Rician) as an extra
    dB term: PL_faded = PL + fading_loss_db where fading_loss_db = -10*log10(|h|^2).
    Fading is deterministic per (tx, rx, epoch). Change 'epoch' to re-sample.
    """
    base: PropagationModel
    kind: str = "rayleigh"     # "rayleigh" or "rician"
    K_dB: float = 6.0          # Rician K-factor (dB) if kind == "rician"
    seed: int = 0
    epoch: Optional[int] = None

    def set_epoch(self, epoch: Optional[int]) -> None:
        self.epoch = epoch

    @staticmethod
    def _rayleigh_gain(rng: random.Random) -> float:
        # |h|^2 ~ Exp(1)
        u = max(rng.random(), 1e-12)
        return -math.log(u)  # exponential(1)

    @staticmethod
    def _rician_gain(rng: random.Random, K_linear: float) -> float:
        # h = s + n, with s = sqrt(K/(K+1)), n ~ CN(0, 1/(K+1))
        sigma = math.sqrt(1.0 / (2.0 * (K_linear + 1.0)))  # per real/imag
        s = math.sqrt(K_linear / (K_linear + 1.0))
        xr = rng.gauss(s, sigma)
        xi = rng.gauss(0.0, sigma)
        return xr * xr + xi * xi  # |h|^2

    def calculate_path_loss(self, tx: WirelessNode, rx: WirelessNode, environment: Environment) -> float:
        pl = self.base.calculate_path_loss(tx, rx, environment)
        rng = random.Random(_pair_seed(self.seed, tx.node_id, rx.node_id, self.epoch))
        if self.kind.lower().startswith("ray"):
            g = self._rayleigh_gain(rng)
        else:
            K_lin = 10 ** (self.K_dB / 10.0)
            g = self._rician_gain(rng, K_lin)
        # add fading loss in dB (negative gain -> positive extra loss)
        fading_loss_db = -10.0 * math.log10(max(g, 1e-12))
        return pl + fading_loss_db




from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Dict
from ..core.digital_twin import DigitalTwin
from .objective import Objective

class Optimizer(ABC):
    @abstractmethod
    def optimize(self, twin: DigitalTwin, objective: Objective) -> Dict[str, Any]:
        raise NotImplementedError



from __future__ import annotations
import math
import numpy as np

try:
    from numba import njit, prange
    HAVE_NUMBA = True
except Exception:  # pragma: no cover
    HAVE_NUMBA = False
    def njit(*args, **kwargs):
        def deco(f): return f
        return deco
    def prange(x): return range(x)

@njit(cache=True, fastmath=True, nogil=True, parallel=True)
def _pairwise_dist_numba(tx_xyz: np.ndarray, rx_xyz: np.ndarray) -> np.ndarray:
    T = tx_xyz.shape[0]
    R = rx_xyz.shape[0]
    out = np.empty((T, R), dtype=np.float64)
    for i in prange(T):
        tx0 = tx_xyz[i, 0]; tx1 = tx_xyz[i, 1]; tx2 = tx_xyz[i, 2]
        for j in range(R):
            dx = tx0 - rx_xyz[j, 0]
            dy = tx1 - rx_xyz[j, 1]
            dz = tx2 - rx_xyz[j, 2]
            out[i, j] = math.sqrt(dx*dx + dy*dy + dz*dz)
    return out

@njit(cache=True, fastmath=True, nogil=True, parallel=True)
def fspl_matrix_db_numba(tx_xyz: np.ndarray, rx_xyz: np.ndarray, f_mhz: float) -> np.ndarray:
    d_m = _pairwise_dist_numba(tx_xyz, rx_xyz)
    T = d_m.shape[0]; R = d_m.shape[1]
    out = np.empty((T, R), dtype=np.float64)
    c2 = 20.0 * math.log10(f_mhz) + 32.44
    for i in prange(T):
        for j in range(R):
            dk = max(d_m[i, j], 1e-3) / 1_000.0
            out[i, j] = 20.0 * math.log10(dk) + c2
    return out

@njit(cache=True, fastmath=True, nogil=True, parallel=True)
def sinr_db_from_rsrp_matrix_numba(
    rsrp_dbm: np.ndarray,  # (T,R)
    serving_tx_idx: np.ndarray,  # (R,)
    noise_dbm: np.ndarray,       # (R,)
) -> np.ndarray:
    T = rsrp_dbm.shape[0]
    R = rsrp_dbm.shape[1]
    out = np.empty(R, dtype=np.float64)
    for j in prange(R):
        # sum powers in mW
        total_mw = 0.0
        for i in range(T):
            total_mw += 10.0 ** (rsrp_dbm[i, j] / 10.0)
        sig_mw = 10.0 ** (rsrp_dbm[serving_tx_idx[j], j] / 10.0)
        noi_mw = 10.0 ** (noise_dbm[j] / 10.0)
        denom = max(total_mw - sig_mw + noi_mw, 1e-30)
        sinr_lin = sig_mw / denom
        out[j] = 10.0 * math.log10(sinr_lin)
    return out



from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.optimization.objective import SumThroughputObjective
from pytwinnet.optimization.placement import PlacementRandomSearchOptimizer, PlacementGridOptimizer

def build_twin():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(300,300,30)))
    twin.set_propagation_model(FreeSpacePathLoss())
    net = Network()
    net.add_node(WirelessNode("gNB-1", position=(50,50,10),
                   transceiver_properties=TransceiverProperties(transmit_power_dbm=32, antenna_gain_dbi=5),
                   metadata={"role":"gNB"}))
    for i, pos in enumerate([(60,70,1.5),(120,120,1.5),(200,210,1.5),(260,240,1.5)]):
        net.add_node(WirelessNode(f"UE-{i+1}", position=pos))
    twin.network = net
    return twin

def main():
    twin = build_twin()
    obj = SumThroughputObjective(tx_id="gNB-1", efficiency=0.75)

    rand_opt = PlacementRandomSearchOptimizer(bounds=((0,300),(0,300)), samples=200, fixed_z=10)
    res_r = rand_opt.optimize(twin, obj, node_ids=["gNB-1"])
    print("Random placement:", res_r["best_positions"], "score=", f"{res_r['best_score']:.2e}")

    grid = [50, 100, 150, 200, 250]
    grid_opt = PlacementGridOptimizer(grid_x=grid, grid_y=grid, fixed_z=10)
    res_g = grid_opt.optimize(twin, obj, node_ids=["gNB-1"])
    print("Grid placement:", res_g["best_positions"], "score=", f"{res_g['best_score']:.2e}")

if __name__ == "__main__":
    main()



from __future__ import annotations
from typing import Tuple, List, Optional
import numpy as np
import matplotlib.pyplot as plt

from ..core.digital_twin import DigitalTwin
from ..physics.batch import path_loss_matrix
from ..accelerate.vectorized import (
    rsrp_matrix_dbm,
    noise_dbm_vector,
    sinr_db_from_rsrp_matrix,
)

def sinr_heatmap_2d_fast(
    twin: DigitalTwin,
    tx_id: str,
    interferer_ids: Optional[List[str]] = None,
    plane_z: float = 1.5,
    xlim: Tuple[float, float] = (0.0, 100.0),
    ylim: Tuple[float, float] = (0.0, 100.0),
    resolution: int = 200,
    bandwidth_hz: float = 20e6,
    noise_figure_db: float = 7.0,
    show: bool = False,
):
    """
    Vectorized SINR heatmap for any PropagationModel.

    - Uses fast, batch path-loss where available.
    - Falls back to a low-overhead scalar loop that reuses node objects.
    """
    assert twin.propagation_model is not None and twin.environment is not None

    tx = twin.network.get_node_by_id(tx_id)
    assert tx is not None, f"Transmitter {tx_id} not found"

    interferers = [twin.network.get_node_by_id(i) for i in (interferer_ids or [])]
    interferers = [n for n in interferers if n is not None]
    txs = [tx] + interferers
    T = len(txs)

    # Grid (R = resolution^2)
    xs = np.linspace(*xlim, resolution)
    ys = np.linspace(*ylim, resolution)
    X, Y = np.meshgrid(xs, ys)
    R = resolution * resolution
    rx_xyz = np.column_stack([X.reshape(-1), Y.reshape(-1), np.full(R, plane_z)])

    # TX arrays for link budget
    tx_dbm = np.array([n.transceiver_properties.transmit_power_dbm for n in txs], dtype=float)
    gt_dbi = np.array([n.transceiver_properties.antenna_gain_dbi for n in txs], dtype=float)
    gr_dbi = np.zeros(R, dtype=float)  # isotropic "probes" on the plane

    # Path-loss matrix (T,R) via generic fast helper
    pl_db = path_loss_matrix(twin, txs, rx_xyz)

    # RSRP per TX-RX
    rsrp_dbm = rsrp_matrix_dbm(tx_dbm, gt_dbi, gr_dbi, pl_db)

    # Assume first TX is serving; the rest are co-channel interferers
    serving_idx = np.zeros(R, dtype=int)
    noise = noise_dbm_vector(bandwidth_hz, noise_figure_db=noise_figure_db) * np.ones(R)
    sinr = sinr_db_from_rsrp_matrix(rsrp_dbm, serving_idx, noise).reshape(resolution, resolution)

    # Plot
    fig, ax = plt.subplots()
    im = ax.imshow(
        sinr, origin="lower",
        extent=(xlim[0], xlim[1], ylim[0], ylim[1]),
        aspect="auto",
    )
    ax.set_xlabel("x (m)"); ax.set_ylabel("y (m)")
    ax.set_title(f"Fast SINR Heatmap (dB) - {tx_id} (+{T-1} interferers)")
    fig.colorbar(im, ax=ax, label="SINR (dB)")
    if show:
        plt.show()
    return ax, sinr



from time import perf_counter
from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.visualization.heatmap_fast import sinr_heatmap_2d_fast

def run(res=600):
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(1000,1000,30)))
    twin.set_propagation_model(FreeSpacePathLoss())
    net = Network()
    net.add_node(WirelessNode("gNB-1",(300,300,10),TransceiverProperties(32,5), metadata={"role":"gNB"}))
    net.add_node(WirelessNode("gNB-2",(700,700,10),TransceiverProperties(32,5), metadata={"role":"gNB"}))
    twin.network = net

    # warm-up (compiles Numba)
    sinr_heatmap_2d_fast(twin, "gNB-1", interferer_ids=["gNB-2"], xlim=(0,1000), ylim=(0,1000), resolution=200)

    t0 = perf_counter()
    _, Z = sinr_heatmap_2d_fast(twin, "gNB-1", interferer_ids=["gNB-2"], xlim=(0,1000), ylim=(0,1000), resolution=res)
    dt = perf_counter() - t0
    print(f"Numba test {res}x{res} -> {Z.shape} in {dt:.3f}s")

if __name__ == "__main__":
    run(300)
    run(600)
    run(900)




from pytwinnet import DigitalTwin, Network, WirelessNode, TransceiverProperties
from pytwinnet.physics import Environment, FreeSpacePathLoss
from pytwinnet.visualization import sinr_heatmap_2d

def main():
    twin = DigitalTwin()
    twin.set_environment(Environment(dimensions_m=(200, 200, 20)))
    twin.set_propagation_model(FreeSpacePathLoss())
    net = Network()
    net.add_node(WirelessNode(node_id="gNB-1", position=(50.0, 50.0, 10.0),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=30.0, antenna_gain_dbi=5.0),
                              metadata={"role":"gNB"}))
    net.add_node(WirelessNode(node_id="gNB-2", position=(150.0, 150.0, 10.0),
                              transceiver_properties=TransceiverProperties(transmit_power_dbm=30.0, antenna_gain_dbi=5.0),
                              metadata={"role":"gNB"}))
    twin.network = net
    sinr_heatmap_2d(twin, tx_id="gNB-1", interferer_ids=["gNB-2"],
                    xlim=(0, 200), ylim=(0, 200), resolution=60, show=False)
    print("Generated SINR heatmap (not displayed in non-GUI env).")

if __name__ == "__main__":
    main()




from __future__ import annotations
from dataclasses import dataclass
from typing import Tuple
from ..core.digital_twin import DigitalTwin

@dataclass(order=True)
class Event:
    timestamp: float
    def apply(self, twin: DigitalTwin) -> None:
        raise NotImplementedError

@dataclass(order=True)
class MoveNodeEvent(Event):
    node_id: str
    new_position: Tuple[float, float, float]
    def apply(self, twin: DigitalTwin) -> None:
        node = twin.network.get_node_by_id(self.node_id)
        if node:
            node.move_to(self.new_position)

@dataclass(order=True)
class TrafficGenerationEvent(Event):
    source_node: str
    dest_node: str
    data_rate_mbps: float
    def apply(self, twin: DigitalTwin) -> None:
        src = twin.network.get_node_by_id(self.source_node)
        dst = twin.network.get_node_by_id(self.dest_node)
        if src and dst:
            src.metadata["generated_traffic_mbps"] = src.metadata.get("generated_traffic_mbps", 0.0) + self.data_rate_mbps
            dst.metadata["received_traffic_mbps"] = dst.metadata.get("received_traffic_mbps", 0.0) + self.data_rate_mbps
